"""Types."""

# pylint: disable=unused-import
from httpx import Response as HTTPResponse  # noqa: F401, pragma: no cover
from httpx._types import (  # noqa: F401, pragma: no cover
    AuthTypes,
    HeaderTypes,
    URLTypes,
)



"""Test utilities here."""
import os
from datetime import datetime, timedelta
from pathlib import PosixPath, WindowsPath
from typing import Any, Dict, Iterable, Union, cast

from _pytest.python_api import ApproxBase

PathClass = PosixPath if os.name == "posix" else WindowsPath


class TmpDir(PathClass):  # type: ignore
    """Extends Path with `cat` and `gen` methods."""

    def cat(self) -> Union[str, Dict[str, Any]]:
        """Returns (potentially multiple) paths' contents.

        Returns:
            a dict of {path: contents} if the path is a directory,
            otherwise the path contents is returned.
        """
        if self.is_dir():
            return {path.name: path.cat() for path in self.iterdir()}
        return cast(str, self.read_text())

    def gen(
        self, struct: Union[str, Dict[str, Any]], text: Union[str, bytes] = ""
    ) -> Iterable[str]:
        """Creates folder structure locally from the provided structure.

        Args:
            root: root of the folder
            struct: the structure to create, can be a dict or a str.
                Dictionary can be nested, which it will create a directory.
                If it's a string, a file with `text` is created.
            text: optional, only necessary if struct is passed a string.
        """
        if isinstance(struct, (str, bytes, PathClass)):
            struct = {struct: text}
        for name, contents in struct.items():
            path = self / name

            if isinstance(contents, dict):
                if not contents:
                    path.mkdir(parents=True, exist_ok=True)
                else:
                    path.gen(contents)
            else:
                path.parent.mkdir(parents=True, exist_ok=True)
                if isinstance(contents, bytes):
                    path.write_bytes(contents)
                else:
                    path.write_text(contents, encoding="utf-8")

        return list(struct.keys())


class approx_datetime(ApproxBase):
    """Perform approximate comparisons between datetime or timedelta.

    See https://github.com/pytest-dev/pytest/issues/8395#issuecomment-790549327
    """

    default_tolerance = timedelta(seconds=1)
    expected: datetime
    abs: timedelta

    def __init__(
        self,
        expected: datetime,
        abs: timedelta = default_tolerance,
    ) -> None:
        """Initialize the approx_datetime with `abs` as tolerance."""
        assert isinstance(expected, datetime)
        assert abs >= timedelta(
            0
        ), f"absolute tolerance can't be negative: {abs}"
        super().__init__(expected, abs=abs)

    def __repr__(self) -> str:  # pragma: no cover
        """String repr for approx_datetime, shown during failure."""
        return f"approx_datetime({self.expected!r} ± {self.abs!r})"

    def __eq__(self, actual) -> bool:
        """Checking for equality with certain amount of tolerance."""
        if isinstance(actual, datetime):
            return abs(self.expected - actual) <= self.abs
        raise AssertionError(  # pragma: no cover
            "expected type of datetime or timedelta"
        )



"""Configuration file for the Sphinx documentation builder."""

# pylint: disable=invalid-name

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
import sys

from pkg_resources import get_distribution

sys.path.insert(0, "..")

project = "webdav4"
copyright = "2021, Saugat Pachhai"  # pylint: disable=redefined-builtin
author = "Saugat Pachhai"

version = get_distribution("webdav4").version
# -- General configuration ---------------------------------------------------

extensions = [
    "myst_parser",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    # external
    "sphinx_copybutton",
]

templates_path = ["_templates"]
todo_include_todos = True

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "fsspec": ("https://filesystem-spec.readthedocs.io/en/stable", None),
}

html_title = "webdav4"
html_static_path = ["_static"]

copybutton_prompt_text = (
    r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True
autoclass_content = "both"



"""Test fixtures."""

from typing import Tuple

import pytest
from cheroot import wsgi

from webdav4.client import Client
from webdav4.fsspec import WebdavFileSystem
from webdav4.urls import URL

from .server import AUTH, get_server_address, run_server
from .utils import TmpDir


@pytest.fixture(autouse=True)
def reduce_backoff_factor():
    """Reduce backoff factor in tests."""
    from webdav4 import retry

    retry.BACKOFF = 0.001


@pytest.fixture
def auth() -> Tuple[str, str]:
    """Auth for the server."""
    return AUTH


@pytest.fixture
def storage_dir(tmp_path_factory) -> TmpDir:
    """Storage for webdav server to keep files in."""
    path = tmp_path_factory.mktemp("webdav")
    return TmpDir(path)


@pytest.fixture
def server(
    storage_dir: TmpDir,
    auth: Tuple[str, str],
) -> wsgi.Server:
    """Creates a server fixture for testing purpose."""
    with run_server("localhost", 0, str(storage_dir), auth) as (httpd, _):
        yield httpd


@pytest.fixture
def server_address(server: wsgi.Server) -> URL:
    """Address of the server to contact."""
    return get_server_address(server)


@pytest.fixture
def client(auth: Tuple[str, str], server_address: URL) -> Client:
    """Webdav client to interact with the server."""
    return Client(server_address, auth=auth)


@pytest.fixture
def fs(client: Client, server_address: URL) -> WebdavFileSystem:
    """Fixture of WebdavFileSystem to interact with the server."""
    return WebdavFileSystem(server_address, client=client)



# pylint: disable=too-many-lines,invalid-name
"""CLI for the Webdav."""

import argparse
import errno
import fileinput
import logging
import os
import posixpath
import re
import sys
from argparse import Namespace
from datetime import datetime, timedelta, timezone
from posixpath import sep
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    cast,
)

from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem
from fsspec.utils import stringify_path

from webdav4 import urls

from .fsspec import WebdavFileSystem
from .urls import URL

if TYPE_CHECKING:
    from argparse import ArgumentParser

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


URL_SCHEME_RE = re.compile(r"^[a-z][a-z0-9.+-]*://", re.IGNORECASE)


class File(NamedTuple):
    """DS that helps us to not compute ext and basename again and again."""

    path: str
    ext: str
    basename: str

    @classmethod
    def from_path(cls, path: str) -> "File":
        """Build the File object from the given path."""
        without_dot = slice(1, None)  # we don't want `.` in the ext
        _, ext = posixpath.splitext(path)
        return cls(path, ext[without_dot], posixpath.basename(path))


def lexical_relpath(path: str, start: str) -> str:
    """Calculates relpath lexically.

    It differs on how it returns the basename for the same paths,
    and bails out if the prefix does not match, as opposed to the
    os.path.relpath.
    """
    if path == start:
        return posixpath.basename(path)
    if start and f"{start}{sep}" in path:
        return path[len(start) + 1 :]
    return path


class LSTheme:
    """Theming support for `ls` command.

    Though, we do use it in other places as well.
    """

    def __init__(
        self, lscolors: str = None, dir_trailing_slash: bool = False
    ) -> None:
        """Build theme from optional lscolors and other configs."""
        self.extensions: Dict[str, str] = {}
        self.codes: Dict[str, str] = {}
        self.dir_trailing_slash: bool = dir_trailing_slash
        self.lscolors: str = lscolors or os.environ.get("LS_COLORS") or ""

        self.colored: bool = self._should_color()
        self._load_ls_colors(self.lscolors)

    @staticmethod
    def _should_color() -> bool:
        """See if we should print in colors or not.

        This honors FORCE_COLOR and NO_COLOR envvars as well. :).
        """
        force_color = os.getenv("FORCE_COLOR")
        if force_color and force_color in ("0", "false"):
            return False
        if force_color:
            return True
        if not sys.stdout.isatty() or os.getenv("TERM") == "dumb":
            return False
        return "NO_COLOR" not in os.environ

    # see:
    # https://github.com/iterative/dvc/blob/1335c7f/dvc/command/ls/ls_colors.py
    def _load_ls_colors(self, lscolors: str) -> None:
        """Loads lscolors."""
        for item in lscolors.split(":"):
            try:
                code, color = item.split("=", 1)
            except ValueError:
                continue
            if code.startswith("*."):
                self.extensions[code[2:]] = color
            else:
                self.codes[code] = color

    def style_path(self, path: str, **info: Any) -> str:
        """Style path in the ls output."""
        isdir: bool = info.get("isdir", False)

        if self.dir_trailing_slash and isdir and not path.endswith(sep):
            path += sep

        if not self.colored:
            return path

        file = File.from_path(path)
        val = None
        if isdir:
            val = self.codes.get("di")
        elif file.ext:
            val = self.extensions.get(file.ext, None)

        if val:
            reset = self.codes.get("rs", 0)
            return f"\033[{val}m{file.path}\033[{reset}m"
        return color_file(file, isdir=isdir)

    def style_size(self, size: str, suff: str = "") -> str:
        """Style size in the ls output, also humanizes it."""
        if suff == "-" and self.colored:
            return size + colored(suff, style="dim")
        if self.colored:
            return colored(size, "green", "bright") + colored(suff, "green")
        return size + suff

    def style_datetime(self, mtime: str) -> str:
        """Humanize and style datetime in the ls output."""
        return colored(mtime, "blue") if self.colored else mtime


theme = LSTheme()


# Ported from exa, with some small tweaks
# https://github.com/ogham/exa/blob/a6754f3cc3/src/info/filetype.rs
def is_build_config_or_readme(file: File) -> bool:
    """is_immediate in terms of exa.

    This file build or generate something (eg: Makefile).
    """
    return file.basename.lower().startswith("readme") or file.basename in {
        "ninja",
        "Makefile",
        "Cargo.toml",
        "SConstruct",
        "CMakeLists.txt",
        "build.gradle",
        "pom.xml",
        "Rakefile",
        "package.json",
        "Gruntfile.js",
        "Gruntfile.coffee",
        "BUILD",
        "BUILD.bazel",
        "WORKSPACE",
        "build.xml",
        "webpack.config.js",
        "meson.build",
        "composer.json",
        "RoboFile.php",
        "PKGBUILD",
        "Justfile",
        "Procfile",
        "Dockerfile",
        "Containerfile",
        "Vagrantfile",
        "Brewfile",
        "Gemfile",
        "Pipfile",
        "build.sbt",
        "mix.exs",
        "bsconfig.json",
        "tsconfig.json",
    }


def is_image(file: File) -> bool:
    """See if the ext is an image type."""
    return file.ext in {
        "png",
        "jpeg",
        "jpg",
        "gif",
        "bmp",
        "tiff",
        "tif",
        "ppm",
        "pgm",
        "pbm",
        "pnm",
        "webp",
        "raw",
        "arw",
        "svg",
        "stl",
        "eps",
        "dvi",
        "ps",
        "cbr",
        "jpf",
        "cbz",
        "xpm",
        "ico",
        "cr2",
        "orf",
        "nef",
        "heif",
    }


def is_video(file: File) -> bool:
    """See if the ext is a Video type."""
    return file.ext in {
        "avi",
        "flv",
        "m2v",
        "m4v",
        "mkv",
        "mov",
        "mp4",
        "mpeg",
        "mpg",
        "ogm",
        "ogv",
        "vob",
        "wmv",
        "webm",
        "m2ts",
        "heic",
    }


def is_music(file: File) -> bool:
    """See if the ext is a Music type."""
    return file.ext in {
        "aac",
        "m4a",
        "mp3",
        "ogg",
        "wma",
        "mka",
        "opus",
        "alac",
        "ape",
        "flac",
        "wav",
    }


def is_crypto(file: File) -> bool:
    """See if the file ext is related to the crypto."""
    return file.ext in {
        "asc",
        "enc",
        "gpg",
        "pgp",
        "sig",
        "signature",
        "pfx",
        "p12",
    }


def is_document(file: File) -> bool:
    """See if the ext is a document type."""
    return file.ext in {
        "djvu",
        "doc",
        "docx",
        "dvi",
        "eml",
        "eps",
        "fotd",
        "key",
        "keynote",
        "numbers",
        "odp",
        "odt",
        "pages",
        "pdf",
        "ppt",
        "pptx",
        "rtf",
        "xls",
        "xlsx",
    }


def is_compressed(file: File) -> bool:
    """See if the ext is a compressed filetype."""
    return file.ext in {
        "zip",
        "tar",
        "Z",
        "z",
        "gz",
        "bz2",
        "a",
        "ar",
        "7z",
        "iso",
        "dmg",
        "tc",
        "rar",
        "par",
        "tgz",
        "xz",
        "txz",
        "lz",
        "tlz",
        "lzma",
        "deb",
        "rpm",
        "zst",
    }


def is_temp(file: File) -> bool:
    """See if the file is a temporary one."""
    return (
        file.basename.endswith("~")
        or (file.basename.startswith("#") and file.basename.endswith("#"))
        or file.ext in {"tmp", "swp", "swo", "swn", "bak", "bk"}
    )


def is_compiled(file: File) -> bool:
    """See if the file is a compiled one."""
    return file.ext in {"class", "elc", "hi", "o", "pyc", "zwc", "ko"}


def _color_file(  # noqa: C901
    file: File, isdir: bool = False
) -> Optional[str]:
    """Returns a color appropriate for file/dir based on ext/path etc."""
    try:
        import colorama  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:  # pragma: no cover
        return None

    color: Optional[str] = None
    if isdir:
        color = colorama.Style.BRIGHT + colorama.Fore.BLUE
    elif is_temp(file):
        color = colorama.Fore.LIGHTWHITE_EX
    elif is_build_config_or_readme(file):
        color = colorama.Fore.YELLOW
    elif is_image(file):
        color = colorama.Fore.CYAN
    elif is_music(file):
        color = colorama.Fore.RED
    elif is_video(file):
        color = colorama.Fore.MAGENTA
    elif is_crypto(file):
        color = colorama.Fore.LIGHTGREEN_EX
    elif is_document(file):
        color = colorama.Fore.BLUE
    elif is_compressed(file):
        color = colorama.Fore.LIGHTRED_EX
    elif is_compiled(file):
        color = colorama.Fore.GREEN

    return color if color else None


def color_file(file: File, isdir: bool = False) -> str:
    """Apply an appropriate color on a file."""
    color = _color_file(file, isdir=isdir)
    if color:
        try:
            import colorama  # pylint: disable=import-outside-toplevel
        except ModuleNotFoundError:  # pragma: no cover
            pass
        else:
            reset = colorama.Style.RESET_ALL
            assert isinstance(reset, str)
            return color + file.path + reset
    return file.path


def colored(name: str, color: str = "", style: str = "") -> str:
    """Colors a string with given color name."""
    try:
        import colorama  # pylint: disable=import-outside-toplevel

        colors = {
            "green": colorama.Fore.GREEN,
            "red": colorama.Fore.RED,
            "blue": colorama.Fore.BLUE,
            "gray": colorama.Fore.LIGHTWHITE_EX,
        }
        styles = {"bright": colorama.Style.BRIGHT, "dim": colorama.Style.DIM}
        reset = colorama.Style.RESET_ALL
    except ModuleNotFoundError:  # pragma: no cover
        colors = {}
        reset = ""
        styles = {}

    _color = colors.get(color, "")
    _style = styles.get(style, "")
    return f"{_style}{_color}{name}{reset}"


def to_fixed_width(n: float, max_width: int) -> str:
    """Converts a float to the fixed max_width."""
    for i in range(max_width - 2, -1, -1):
        float_str = f"{n:.{i}f}"
        if len(float_str) <= max_width:
            break
    assert max_width >= 2  # pragma: no cover
    return float_str


def human_size(nbytes: Optional[float]) -> Tuple[str, str]:
    """Converts bytes to human-readable size."""
    if nbytes is None:
        return "", "-"
    if nbytes < 1024:
        return str(nbytes), ""

    suffixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]
    for suff in suffixes:  # noqa: B007, pragma: no cover
        if nbytes < 1024:
            break
        nbytes /= 1024

    return to_fixed_width(nbytes, 3), suff


def format_datetime(mtime: Any) -> str:
    """Converts mtime to ls-compatible output."""
    if not isinstance(mtime, datetime):
        return "-"
    fmt = "%b %d %H:%M"
    # it's mostly for presentation, so we don't care much about tz
    if mtime.replace(tzinfo=None) < datetime.today() - timedelta(days=180):
        fmt = "%b %d %Y"
    return mtime.strftime(fmt)


def process_url(url: str) -> str:
    """Support url without schemes and the shorthand format for localhost."""
    if not URL_SCHEME_RE.match(url):
        # See if we're using curl style shorthand for localhost (:3000/foo)
        shorthand = re.match(r"^:(?!:)(\d*)(/?.*)$", url)
        if shorthand:
            port = shorthand.group(1)
            rest = shorthand.group(2)
            url = "http://" + "localhost"
            if port:
                url += ":" + port
            url += rest
        else:
            url = "https://" + url
    return url


def prepare_url_auth(  # noqa: C901
    args: Namespace,
) -> Tuple[urls.URL, Optional[Tuple[str, str]]]:
    """Process url and auth from the given arguments or from the envvar.

    That includes using user and password from url itself.
    """
    url = args.endpoint_url
    if not url:
        url = os.getenv("WEBDAV_ENDPOINT_URL")

    if not url:
        raise ValueError(
            "no endpoint url specified, "
            "please specify it through --endpoint-url "
            "or via WEBDAV_ENDPOINT_URL envvar."
        )

    url_obj = URL(process_url(url))
    user, password = None, None
    auth = None

    if url_obj.username:
        user = url_obj.username
    elif args.user:
        user = args.user

    if url_obj.password:
        password = url_obj.password
    elif args.password:
        password = args.password

    if user and password:
        auth = user, password

    return url_obj, auth


class Command:
    """Base class for all commands."""

    def __init__(self, args: Namespace, fs: AbstractFileSystem = None) -> None:
        """Pass the arguments and optionally fs."""
        self.args = args
        self.auth = None
        self.endpoint_url = None
        if fs:
            self.fs = fs
            return

        self.endpoint_url, self.auth = prepare_url_auth(args)
        logger.debug("base url set to %s", self.endpoint_url)
        self.fs = WebdavFileSystem(self.endpoint_url, auth=self.auth)

    def run(self) -> None:
        """Override this function to do some operations."""
        raise NotImplementedError


class Size(NamedTuple):
    """Size data structure carrying suffix and bytes."""

    nbytes: str
    suff: str = ""


class Row(NamedTuple):
    """Row data structure representing each row for the ls output."""

    date: str
    size: Size
    file: str
    isdir: bool = False


class CommandLS(Command):
    """Command for ls."""

    def ls(self) -> Iterator[Row]:
        """Returns a list of rows that are styled."""
        # if user set a `-L`, we would still like to show
        # directory in L level.
        withdirs = not self.args.recursive or bool(self.args.level)
        depth = self.args.level if self.args.recursive else 1
        details = self.fs.find(
            self.args.path, maxdepth=depth, detail=True, withdirs=withdirs
        )

        path = self.fs._strip_protocol(  # pylint: disable=protected-access
            self.args.path
        ).strip("/")
        path_level = path.count(sep) + 1 if path else 0

        if not details:
            # if there are no information, it's likely that
            # the file does not exist at all.
            self.fs.info(path)
        elif len(details) == 1 and self.fs.isfile(path):
            # or, the entry was a file, so fsspec did not return
            # details for it.
            details = {path: self.fs.info(path)}

        for file_path, info in details.items():
            isdir = info.get("type") == "directory"
            file_path = file_path.strip("/")
            within_depth = depth and (
                0 <= file_path.count(sep) - path_level < depth - 1
            )

            if withdirs and isdir and within_depth:
                continue

            file_name = file_path
            if not self.args.full_path:
                file_name = lexical_relpath(file_path, path)

            date = format_datetime(info.get("modified"))
            size = Size(*human_size(info.get("size")))
            yield Row(date=date, size=size, file=file_name, isdir=isdir)

    @staticmethod
    def render(details: List[Row]) -> None:
        """Display provided information in a columnar format."""
        if not details:
            return None

        date_maxsize = max(len(row.date) for row in details)
        size_maxsize = max(len(row.size.nbytes) for row in details)

        for row in details:
            date = theme.style_datetime(row.date.ljust(date_maxsize))

            size_just = size_maxsize if row.size.suff else size_maxsize + 1
            size = theme.style_size(
                row.size.nbytes.rjust(size_just), suff=row.size.suff
            )
            file = theme.style_path(row.file, isdir=row.isdir)
            print(date, size, file)

        return None

    def run(self) -> None:
        """List and render the ls output."""
        try:
            details = list(self.ls())
        except FileNotFoundError as exc:
            raise FileNotFoundError(
                errno.ENOENT, "No such file or directory", self.args.path
            ) from exc
        return self.render(details)


class CommandTransfer(Command):
    """Shared class for move and copy."""

    def transfer_remote(self) -> None:
        """Transfer between remote-remote operations."""
        raise NotImplementedError

    def run(self) -> None:
        """Run transfer between two paths."""
        path1 = self.args.path1
        path2 = self.args.path2
        recursive = self.args.recursive

        if is_fqpath(self.fs, path1) and is_fqpath(self.fs, path2):
            self.transfer_remote()
        elif is_fqpath(self.fs, path1):
            if self.fs.isdir(path1):
                os.makedirs(path2, exist_ok=True)
            self.fs.download(path1, path2, recursive=recursive)
        else:
            self.fs.upload(path1, path2, recursive=recursive)


class CommandCopy(CommandTransfer):
    """Command for cp."""

    def transfer_remote(self) -> None:
        """Copy files/diretories between remotes."""
        self.fs.cp(
            self.args.path1, self.args.path2, recursive=self.args.recursive
        )


class CommandMove(CommandTransfer):
    """Command for mv.

    Does not remove src, is similar to cp for now.
    """

    def transfer_remote(self) -> None:
        """Move files/diretories between remotes."""
        self.fs.mv(
            self.args.path1, self.args.path2, recursive=self.args.recursive
        )


class CommandRemove(Command):
    """Command for rm."""

    def run(self) -> None:
        """Remove files/directories in the remote."""
        details = self.fs.info(self.args.path)
        if details["type"] == "directory" and not self.args.recursive:
            self.fs.rmdir(self.args.path)
        else:
            self.fs.rm(self.args.path, recursive=self.args.recursive)


class CommandRun(Command):
    """Command for run."""

    def run(self) -> None:
        """Runs the command from stdin/file."""
        if not sys.stdout.isatty() and not self.args.path:
            raise ValueError("no path specified or contents piped")

        count = 0
        for line in fileinput.input(files=(self.args.path or "-",)):
            if line.startswith("#"):
                continue

            cmd, *args = line.strip().split()
            subparsers = self.args.subparsers
            if cmd not in subparsers:
                raise ValueError(f"unknown command: {cmd}")

            subparser = subparsers[cmd]
            sub_args = subparser.parse_args(args)

            namespace = Namespace(**vars(self.args))
            namespace.__dict__.update(vars(sub_args))
            logger.debug(str(vars(namespace)))

            print("\n" if count else "", line, sep="", end="")
            run_cmd(namespace, self.fs)
            count += 1


class CommandCat(Command):
    """Command for cat."""

    def run(self) -> None:
        """Prints the content of the file."""
        with self.fs.open(self.args.path, mode="r") as fobj:
            print(fobj.read())


def is_fqpath(fs: AbstractFileSystem, path: str) -> bool:
    """Check if the path is fully qualified."""
    path = stringify_path(path)
    protos = (fs.protocol,) if isinstance(fs.protocol, str) else fs.protocol
    for protocol in protos:
        if path.startswith(protocol + "://"):
            return True
    return False


class CommandSync(Command):
    """Command for sync."""

    @staticmethod
    def copy_fs(
        src: str,
        dest: str,
        src_fs: AbstractFileSystem,
        dest_fs: AbstractFileSystem,
        recursive: bool = False,
    ) -> None:
        """Supports copy in local-remote, remote-local, and remote-remote."""
        assert not (
            isinstance(src_fs, LocalFileSystem)
            and isinstance(dest_fs, LocalFileSystem)
        )

        if isinstance(src_fs, LocalFileSystem):
            dest_fs.upload(src, dest, recursive=recursive)
        elif isinstance(dest_fs, LocalFileSystem):
            if src_fs.isdir(src):
                dest_fs.makedirs(dest, exist_ok=True)
            src_fs.download(src, dest, recursive=recursive)
        else:
            dest_fs.copy(src, dest, recursive=recursive)

    @classmethod  # noqa: C901
    def changed(  # noqa: C901
        cls, src_details: Dict[str, Any], dest_details: Dict[str, Any]
    ) -> bool:
        """See if a src and dest have changed or not."""

        def get_mtime(info: Dict[str, Any]) -> Optional[datetime]:
            mtime = (
                info.get("mtime")
                or info.get("modified")
                or info.get("created")
            )
            if isinstance(mtime, float):
                return datetime.utcfromtimestamp(mtime).replace(
                    tzinfo=timezone.utc
                )
            return mtime

        src_type = src_details.get("type")
        src_size = src_details.get("size")
        src_modified = get_mtime(src_details)

        dest_type = dest_details.get("type")
        dest_size = dest_details.get("size")
        dest_modified = get_mtime(dest_details)

        if not dest_details:
            return True
        if src_type == dest_type == "file":
            if src_size != dest_size:
                return True
            if src_modified and dest_modified and src_modified > dest_modified:
                return True
        if src_type == dest_type == "directory":
            return True
        if src_type != dest_type:
            raise TypeError(
                "cannot sync between different types, "
                f"src is {src_type}, "
                f"dest is {dest_type}"
            )
        return False

    @classmethod
    def diff(
        cls,
        src_info: Dict[str, Dict[str, Any]],
        dest_info: Dict[str, Dict[str, Any]],
    ) -> Tuple[List[str], Set[str]]:
        """Diff between src info and dest info to see if they have changed."""
        only_in_dest: Set[str] = set(dest_info) - set(src_info)
        changed: List[str] = []
        for file, src_d in src_info.items():
            dest_d = dest_info.get(file)
            if cls.changed(src_d, dest_d or {}):
                changed.append(file)

        return changed, only_in_dest

    @staticmethod
    def _transform_info(info: Dict[str, Any], rel: str) -> Dict[str, Any]:
        """Convert to relative paths for easier diff comparison."""
        return {os.path.relpath(k, rel): v for k, v in info.items()}

    def sync(
        self,
        src: str,
        dest: str,
        src_fs: AbstractFileSystem,
        dest_fs: AbstractFileSystem,
    ) -> None:
        """Sync between src and dest."""
        details_src = src_fs.info(src)
        try:
            details_dest = dest_fs.info(dest)
        except IOError:
            isdir = details_src.get("type") == "directory"
            print(
                "copy:",
                theme.style_path(src, isdir=isdir),
                "to",
                theme.style_path(dest, isdir=isdir),
            )
            self.copy_fs(src, dest, src_fs, dest_fs, recursive=True)
            return None

        if details_src["type"] == "file" and self.changed(
            details_src, details_dest
        ):
            print("copy:", theme.style_path(src), "to", theme.style_path(dest))
            self.copy_fs(src, dest, src_fs, dest_fs, recursive=False)
            return None

        src_info = src_fs.find(src, withdirs=True, detail=True, maxdepth=1)
        dest_info = dest_fs.find(dest, withdirs=True, detail=True, maxdepth=1)

        src_d = self._transform_info(src_info, src)
        dest_d = self._transform_info(dest_info, dest)
        changed, only_in_dest = self.diff(src_d, dest_d)

        for file in changed:
            new_src = posixpath.join(src, file)
            new_dest = posixpath.join(dest, file)
            self.sync(new_src, new_dest, src_fs, dest_fs)

        for file in only_in_dest if self.args.delete else []:
            new_dest = posixpath.join(dest, file)
            is_dir = dest_d[file].get("type") == "directory"
            print("delete:", theme.style_path(new_dest, isdir=is_dir))
            dest_fs.rm(new_dest, recursive=True)
        return None

    def run(self) -> None:
        """Run sync between src and dest."""
        src = self.args.path1
        dest = self.args.path2

        src_is_fq = is_fqpath(self.fs, src)
        dest_is_fq = is_fqpath(self.fs, dest)

        if src_is_fq and dest_is_fq:
            src_fs, dest_fs = self.fs, self.fs
        elif src_is_fq:
            src_fs, dest_fs = self.fs, LocalFileSystem()
        else:
            src_fs, dest_fs = LocalFileSystem(), self.fs

        # pylint: disable=protected-access
        src = src_fs._strip_protocol(src)
        dest = dest_fs._strip_protocol(dest)
        return self.sync(src, dest, src_fs, dest_fs)


class CommandMkdir(Command):
    """Command for mkdir."""

    def run(self) -> None:
        """Create directory."""
        self.fs.mkdir(self.args.path, create_parents=self.args.parents)


class CommandDiskUsage(Command):
    """Command for du."""

    def run(self) -> None:
        """Displays disk usages for the given path."""
        path = self.args.path
        isdir = True
        details = self.fs.find(path, detail=True, withdirs=False)
        if len(details) == 1 and self.fs.isfile(path):
            isdir = False
            details = {path: self.fs.info(path)}

        size = sum(info.get("size", 0) for _, info in details.items())
        sized = Size(*human_size(size))
        size_styled = theme.style_size(sized.nbytes, suff=sized.suff)
        path_styled = theme.style_path(path, isdir=isdir)
        files_str = "files" if len(details) > 1 else "file"
        print(
            size_styled,
            f"({size} bytes) in",
            len(details),
            f"{files_str}:",
            path_styled,
        )


def get_parser() -> Tuple["ArgumentParser", Dict[str, "ArgumentParser"]]:
    """Returns the parser and dict of subparsers.

    We use this dict of subparsers to parse commands in `run` command.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Show more information",
        default=False,
    )
    parser.add_argument(
        "--endpoint-url",
        help="Endpoint url to connect to. Will be considered as a root path.\n"
        "Can also be specified through WEBDAV_ENDPOINT_URL envvar.",
        metavar="URL",
        default=None,
    )
    parser.add_argument(
        "--user", "-u", help="Account Username", default=None, required=False
    )
    parser.add_argument(
        "--password",
        "-p",
        help="Account Password",
        default=None,
        required=False,
    )

    subparsers = parser.add_subparsers(
        title="actions", help="Available subcommands"
    )
    subparsers.required = True
    subparsers.dest = "command"

    ls_parser = subparsers.add_parser("ls", help="List files")
    ls_parser.add_argument(
        "--recursive",
        "-R",
        default=False,
        action="store_true",
        help="Recurse into directories",
    )
    ls_parser.add_argument(
        "-L",
        "--level",
        required=False,
        default=None,
        type=int,
        help="Limit the depth of recursion",
    )
    ls_parser.add_argument(
        "--full-path",
        required=False,
        action="store_true",
        help="Show full path of the files and the directories",
    )
    ls_parser.add_argument(
        "path", nargs="?", default="dav://", help="Path to list from"
    )
    ls_parser.set_defaults(func=CommandLS)

    cp_parser = subparsers.add_parser(
        "cp",
        help="Copies a local file or remote file "
        "to another location locally or in remote.",
    )
    cp_parser.add_argument(
        "--recursive", "-R", default=False, action="store_true", help=""
    )
    cp_parser.add_argument(
        "path1",
        help="Path to copy from (locally or remotely if dav:// url is given)",
    )
    cp_parser.add_argument(
        "path2",
        help="Path to copy to (locally or remotely if dav:// url is given)",
    )
    cp_parser.set_defaults(func=CommandCopy)

    mv_parser = subparsers.add_parser(
        "mv",
        help="Moves a local file or remote file "
        "to another location locally or in remote.",
    )
    mv_parser.add_argument(
        "--recursive", "-R", default=False, action="store_true", help=""
    )
    mv_parser.add_argument(
        "path1",
        help="Path to move from (locally or remotely if dav:// url is given)",
    )
    mv_parser.add_argument(
        "path2",
        help="Path to move to (locally or remotely if dav:// url is given)",
    )
    mv_parser.set_defaults(func=CommandMove)

    rm_parser = subparsers.add_parser(
        "rm", help="Removes a file from the remote server."
    )
    rm_parser.add_argument(
        "--recursive", "-R", default=False, action="store_true", help=""
    )
    rm_parser.add_argument("path", help="Path to remove")
    rm_parser.set_defaults(func=CommandRemove)

    mkdir_parser = subparsers.add_parser(
        "mkdir", help="Creates a directory/collection in the remote server."
    )
    mkdir_parser.add_argument(
        "--parents",
        "-p",
        default=False,
        action="store_true",
        help="no error if existing, " "make parent directories as needed",
    )
    mkdir_parser.add_argument("path", help="Path to remove")
    mkdir_parser.set_defaults(func=CommandMkdir)

    run_parser = subparsers.add_parser("run", help="Run multiple commands")
    run_parser.add_argument(
        "path", metavar="FILE", help="files to read, if empty, stdin is used"
    )
    run_parser.set_defaults(func=CommandRun)

    cat_parser = subparsers.add_parser("cat", help="Print remote file content")
    cat_parser.add_argument("path", metavar="FILE", help="File to read")
    cat_parser.set_defaults(func=CommandCat)

    sync_parser = subparsers.add_parser(
        "sync",
        help="Sync local file/folder to/from the remote file/directories",
    )
    sync_parser.add_argument(
        "path1",
        help="Path to move from (locally or remotely if dav:// url is given)",
    )
    sync_parser.add_argument(
        "path2",
        help="Path to move to (locally or remotely if dav:// url is given)",
    )
    sync_parser.add_argument(
        "--delete",
        action="store_true",
        default=False,
        help="Files that exist in the destination "
        "but not in the source are deleted during sync.",
    )
    sync_parser.set_defaults(func=CommandSync)

    du_parser = subparsers.add_parser("du", help="Print remote file content")
    du_parser.add_argument(
        "path",
        default="dav://",
        nargs="?",
        help="Show size of directories/files",
    )
    du_parser.set_defaults(func=CommandDiskUsage)

    return parser, {
        "ls": ls_parser,
        "cp": cp_parser,
        "mv": mv_parser,
        "rm": rm_parser,
        "cat": cat_parser,
        "sync": sync_parser,
        "du": du_parser,
        "mkdir": mkdir_parser,
    }


def run_cmd(args: Namespace, fs: AbstractFileSystem = None) -> Optional[int]:
    """Run cmd from given args."""
    cmd = cast(Command, args.func(args, fs=fs))
    cmd.run()
    return 0


def main(argv: List[str] = None) -> Optional[int]:
    """Command line entrypoint."""
    parser, subparsers = get_parser()
    args = parser.parse_args(argv)
    args.subparsers = subparsers

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    try:
        return run_cmd(args)
    except Exception as exc:  # pylint: disable=broad-except
        logger.error("%s: %s", type(exc).__name__, exc, exc_info=args.verbose)
        return 1



"""Server for testing purposes, used on repl and pytest as a fixture."""

import threading
from contextlib import contextmanager
from typing import ContextManager, Iterator, Tuple

from cheroot import wsgi
from httpx import URL
from wsgidav.wsgidav_app import WsgiDAVApp

AUTH = "user1", "password1"


def get_server_address(srvr: wsgi.Server) -> URL:
    """Returns base URL of the server."""
    return get_url_from_addr("localhost", srvr.bind_addr[1])


def get_url_from_addr(host, port) -> URL:
    """Builds URL from the host and port."""
    return URL("http://{0}:{1}".format(host, port))


@contextmanager
def run_server_on_thread(
    srvr: wsgi.Server,
) -> Iterator[Tuple[wsgi.Server, threading.Thread]]:
    """Runs server on a separate thread."""
    srvr.prepare()
    thread = threading.Thread(target=srvr.serve)
    thread.daemon = True
    thread.start()

    try:
        yield srvr, thread
    finally:
        srvr.stop()
    thread.join()


def run_server(
    host: str,
    port: int,
    directory: str,
    authentication: Tuple[str, str],
) -> ContextManager[Tuple[wsgi.Server, threading.Thread]]:
    """Runs a webdav server."""
    dirmap = {"/": directory}

    user, pwd = authentication
    app = WsgiDAVApp(
        {
            "host": host,
            "port": port,
            "provider_mapping": dirmap,
            "simple_dc": {"user_mapping": {"*": {user: {"password": pwd}}}},
        }
    )
    return run_server_on_thread(
        wsgi.Server(bind_addr=(host, port), wsgi_app=app)
    )


if __name__ == "__main__":
    # usage: python -m tests.server

    import argparse
    import code
    import tempfile
    from contextlib import suppress

    from webdav4.client import Client
    from webdav4.fsspec import WebdavFileSystem

    from .utils import TmpDir

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--interactive",
        "-i",
        help="Run test server interactively",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()
    storage_dir = TmpDir(tempfile.mkdtemp("repl"))

    with run_server("localhost", 0, str(storage_dir), AUTH) as (server, th):
        server_address = get_server_address(server)
        client = Client(server_address, auth=AUTH)
        fs = WebdavFileSystem(server_address, client=client)

        print(f"Running server on {server_address} ...")
        print("Authenticate with {0}:{1}".format(*AUTH))

        try:
            if args.interactive:
                print(
                    "fs, client, server_address, storage_dir fixtures "
                    "are available"
                )
                try:
                    from IPython import embed

                    embed(colors="neutral")
                except ImportError:
                    with suppress(ImportError):
                        import readline  # noqa: F401

                    shell = code.InteractiveConsole({**globals(), **locals()})
                    shell.interact()
            else:
                th.join()
        except KeyboardInterrupt:
            print("Exiting ...")



"""Testing date parsing logic here."""

from datetime import datetime

import pytest as pytest
from dateutil.tz import tzutc

from webdav4.date_utils import from_rfc1123, fromisoformat


def test_iso8601_parsing():
    """Test parsing iso8601 format.

    creationdate is in this format.
    """
    dt = datetime(2020, 1, 2, 3, 4, 5, tzinfo=tzutc())
    datestring = "2020-1-02T03:04:05Z"
    assert fromisoformat(datestring) == dt


def test_rfc1123_parsing():
    """Test parsing rfc1123 format datetime.

    getlastmodified is in this format.
    """
    dt = datetime(2020, 1, 2, 3, 4, 5, tzinfo=tzutc())
    datestring = "Fri, 02 Jan 2020 03:04:05 GMT"
    assert from_rfc1123(datestring) == dt


@pytest.mark.parametrize(
    "datestring",
    [
        "Fri, 2 Jan 2020 03:04:05 GMT",  # non-zero format in the day
        "Fri, 02 Jan 2020 03:04:05 +0000",  # rfc1123z
        "Fri Jan 02 03:04:05 UTC 2020",  # unix date
        "2020-01-02T03:04:05+0000",  # rfc3339
    ],
)
def test_non_standard_datetimes(datestring):
    """Test that we can still parse non-standard datetimes."""
    dt = datetime(2020, 1, 2, 3, 4, 5, tzinfo=tzutc())
    assert fromisoformat(datestring) == dt
    assert from_rfc1123(datestring) == dt



"""Testing fsspec based WebdavFileSystem."""
import errno
from datetime import datetime, timezone
from typing import Any, Dict, List, Set, Tuple, Union

import fsspec
import pytest

from webdav4.fsspec import WebdavFileSystem
from webdav4.urls import URL, join_url

from .test_callback import ReadWrapper
from .utils import TmpDir, approx_datetime


def test_fs_ls(storage_dir: TmpDir, fs: WebdavFileSystem, server_address: URL):
    """Tests fsspec for webdav."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    stat = (storage_dir / "data").stat()

    assert fs.ls("/") == [
        {
            "size": None,
            "created": approx_datetime(
                datetime.fromtimestamp(int(stat.st_ctime), tz=timezone.utc)
            ),
            "modified": approx_datetime(
                datetime.fromtimestamp(int(stat.st_mtime), tz=timezone.utc)
            ),
            "content_language": None,
            "content_type": None,
            "etag": None,
            "type": "directory",
            "name": "data",
            "display_name": "data",
            "href": join_url(server_address, "data").path + "/",
        }
    ]
    foo_stat = (storage_dir / "data" / "foo").stat()
    bar_stat = (storage_dir / "data" / "bar").stat()

    info = fs.ls("/data/")
    assert len(info) == 2

    d = {}
    for i in info:
        assert not isinstance(i, str)
        i.pop("etag", None)
        d[i["name"]] = i

    assert d["data/bar"] == {
        "name": "data/bar",
        "href": join_url(server_address, "data/bar").path,
        "size": 3,
        "created": approx_datetime(
            datetime.fromtimestamp(int(bar_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(bar_stat.st_ctime), tz=timezone.utc)
        ),
        "content_language": None,
        "content_type": "application/octet-stream",
        "type": "file",
        "display_name": "bar",
    }
    assert d["data/foo"] == {
        "name": "data/foo",
        "href": join_url(server_address, "data/foo").path,
        "size": 3,
        "created": approx_datetime(
            datetime.fromtimestamp(int(foo_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(foo_stat.st_ctime), tz=timezone.utc)
        ),
        "display_name": "foo",
        "content_language": None,
        "content_type": "application/octet-stream",
        "type": "file",
    }

    fs.rm("/data/foo")
    assert fs.ls("/data/", detail=False) == ["data/bar"]
    assert fs.size("/data/bar") == 3
    assert fs.modified("/data/bar") == approx_datetime(
        datetime.fromtimestamp(int(stat.st_mtime), tz=timezone.utc)
    )
    assert fs.cat("/data/bar") == b"bar"

    checksum = fs.checksum("data/bar/")
    assert checksum and isinstance(checksum, str)

    fs.mv("data/bar", "data/foobar")
    assert fs.ls("data", detail=False) == ["data/foobar"]

    foobar_stat = (storage_dir / "data" / "foobar").stat()
    assert fs.created("data/foobar") == approx_datetime(
        datetime.fromtimestamp(int(foobar_stat.st_ctime), tz=timezone.utc)
    )

    fs.cp("data/foobar", "data/bar")
    assert set(fs.ls("data", detail=False)) == {"data/foobar", "data/bar"}

    fs.makedirs("data/subdir/subsubdir", exist_ok=True)
    assert fs.isdir("data/subdir/subsubdir")

    fs.mkdir("data/subdir2", create_parents=False)
    assert fs.isdir("data/subdir2")

    fs.mkdir("data/subdir2/subdir3/subdir4")
    assert fs.isdir("data/subdir2/subdir3/subdir4")

    with pytest.raises(FileNotFoundError):
        fs.ls("data/not-existing-file")


def test_open(storage_dir: TmpDir, fs: WebdavFileSystem, server_address: URL):
    """Test opening a remote file from webdav."""
    storage_dir.gen({"data": {"foo": "foo"}})

    with pytest.raises(IsADirectoryError):
        fs.open("data")

    with pytest.raises(FileNotFoundError):
        fs.open("data/bar")

    with fs.open("/data/foo") as f:
        assert f.read() == b"foo"
        assert f.read() == b""
        assert f.read() == b""

    with fs.open("/data/foo", mode="r") as f:
        assert f.read() == "foo"
        assert f.read() == ""
        assert f.read() == ""

    stat = (storage_dir / "data" / "foo").stat()
    with fs.open("/data/foo", mode="r", size=stat.st_size) as f:
        assert f.read() == "foo"
        assert f.read() == ""
        assert f.read() == ""

    fobj = fs.open("/data/foo")

    assert not fobj.closed
    fobj.close()
    assert fobj.closed
    fobj.close()
    assert fobj.closed

    fs.put_file(storage_dir / "data" / "foo", "dir/somewhere/data2")
    assert fs.cat("dir/somewhere/data2") == b"foo"

    fs.put_file(storage_dir / "data", "dir/somewhere2")
    assert fs.isdir("dir/somewhere2")

    fs.pipe_file("dir/somewhere2/foo", b"foo")
    assert fs.cat("dir/somewhere2/foo") == b"foo"

    fs.touch("dir/file")
    assert fs.cat("dir/file") == b""

    fs.touch("dir/somewhere2/foo")
    assert fs.cat("dir/somewhere2/foo") == b""


def test_open_write_bytes_simple(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test simple functions from fileobj in wb mode."""
    file = fs.open("foo", "wb")
    assert not file.closed

    with pytest.raises(ValueError):
        file.info()

    assert file.fileno()
    assert not file.isatty()
    assert file.readable()
    assert file.writable()
    assert file.seekable()

    file.close()
    assert file.closed

    assert storage_dir.cat() == {"foo": ""}


def test_open_write_bytes_write(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test write functions from fileobj in wb mode."""
    with fs.open("foo", "wb") as f:
        assert f.write(b"foo\n") == 4
        assert f.write(b"bar\n") == 4
        assert f.write(b"foobar\n") == 7
        f.writelines([b"lorem\n", b"ipsum\n"])

        assert f.tell() == 27
        f.flush()

    assert storage_dir.cat() == {"foo": "foo\nbar\nfoobar\nlorem\nipsum\n"}

    with fs.open("foo", "wb") as f:
        assert f.write(b"foo\n") == 4
        f.truncate(3)
    assert storage_dir.cat() == {"foo": "foo"}


def test_open_write_bytes_read(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test read functions from fileobj in wb mode."""
    with fs.open("foo", "wb") as f:
        assert f.write(b"foo\n") == 4
        assert f.write(b"bar\n") == 4
        assert f.write(b"foobar\n") == 7
        f.seek(0)
        assert f.read() == b"foo\nbar\nfoobar\n"
        f.seek(10)
        assert f.read(2) == b"ob"
        assert f.read() == b"ar\n"
        f.seek(0)
        assert f.readlines() == [b"foo\n", b"bar\n", b"foobar\n"]
        f.seek(0)
        assert f.readline() == b"foo\n"
        assert f.readuntil(b"o") == b"bar\nfo"
        b = bytearray(5)
        f.readinto(b)
        assert bytes(b) == b"obar\n"
        assert f.tell() == 15

    assert storage_dir.cat() == {"foo": "foo\nbar\nfoobar\n"}


def test_open_write_bytes_commit_discard(
    storage_dir: TmpDir, fs: WebdavFileSystem
):
    """Test commit and discard functions from fileobj in wb mode."""
    with fs.open("foo", "wb") as f:
        f.write(b"hello")
        f.discard()

    assert storage_dir.cat() == {}

    with fs.open("foo", "wb") as f:
        f.write(b"foo")
        f.commit()

    f.discard()
    f.close()
    assert storage_dir.cat() == {"foo": "foo"}


def test_open_write_bytes_x_mode(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test opening file in x mode."""
    with fs.open("foo", "x") as f:
        f.write("foo")

    assert storage_dir.cat() == {"foo": "foo"}

    with pytest.raises(FileExistsError):
        fs.open("foo", "x")


def test_open_write_in_text_mode(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test few methods from fileobj opened in text mode."""
    with fs.open("foo", "w") as f:
        f.write("foobar")
        f.writelines(["lorem"])

        pos = f.tell()
        f.seek(0)
        assert f.read() == "foobarlorem"
        f.seek(pos)
        f.flush()

    assert storage_dir.cat() == {"foo": "foobarlorem"}


def test_exists(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that exists complies with fsspec."""
    storage_dir.gen(
        {"data": {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}}
    )

    assert fs.exists("data")
    assert fs.exists("data/foo")
    assert fs.exists("data/bar")
    assert fs.exists("data/baz")
    assert fs.exists("data/baz/foobaz")
    assert not fs.exists("data2")
    assert not fs.exists("data/bazz")


def test_isdir(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that isdir complies with fsspec."""
    storage_dir.gen(
        {"data": {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}}
    )

    assert fs.isdir("data")
    assert not fs.isdir("data/foo")
    assert not fs.isdir("data/bar")
    assert fs.isdir("data/baz")
    assert not fs.isdir("data/baz/foobaz")

    # not existing ones should return False, instead of just failing
    assert not fs.isdir("data2")
    assert not fs.isdir("data/bazz")


def test_isfile(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that isfile complies with fsspec."""
    storage_dir.gen(
        {"data": {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}}
    )

    assert not fs.isfile("data")
    assert fs.isfile("data/foo")
    assert fs.isfile("data/bar")
    assert not fs.isfile("data/baz")
    assert fs.isfile("data/baz/foobaz")

    # not existing ones should return False, instead of just failing
    assert not fs.isfile("data2")
    assert not fs.isfile("data/bazz")


def test_created(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that created complies with fsspec."""
    with pytest.raises(FileNotFoundError):
        fs.created("not-existing-file")

    storage_dir.gen({"data": {"foo": "foo"}})

    data_stat = (storage_dir / "data" / "foo").stat()
    assert fs.created("data") == approx_datetime(
        datetime.fromtimestamp(int(data_stat.st_ctime), tz=timezone.utc)
    )

    foo_stat = (storage_dir / "data" / "foo").stat()
    assert fs.created("data/foo") == approx_datetime(
        datetime.fromtimestamp(int(foo_stat.st_ctime), tz=timezone.utc)
    )


def test_modified(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that modified complies with fsspec."""
    with pytest.raises(FileNotFoundError):
        fs.created("not-existing-file")

    storage_dir.gen({"data": {"foo": "foo"}})

    data_stat = (storage_dir / "data" / "foo").stat()
    assert fs.modified("data") == approx_datetime(
        datetime.fromtimestamp(int(data_stat.st_mtime), tz=timezone.utc)
    )

    foo_stat = (storage_dir / "data" / "foo").stat()
    assert fs.modified("data/foo") == approx_datetime(
        datetime.fromtimestamp(int(foo_stat.st_mtime), tz=timezone.utc)
    )


def test_checksum(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that checksum complies with fsspec."""
    with pytest.raises(FileNotFoundError):
        fs.created("not-existing-file")

    storage_dir.gen({"data": {"foo": "foo"}})
    assert fs.checksum("data") is None  # is a directory
    assert fs.checksum("data/foo")


def test_size(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that size complies with fsspec."""
    with pytest.raises(FileNotFoundError):
        fs.created("not-existing-file")

    storage_dir.gen({"data": {"foo": "foo"}})
    assert fs.size("data") is None  # is a directory
    assert fs.size("data/foo") == 3


def test_disk_usage(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that du complies with fsspec."""
    with pytest.raises(FileNotFoundError):
        fs.created("not-existing-file")

    storage_dir.gen({"data": {"foo": "foo"}})
    assert fs.du("data") == 3
    assert fs.du("data/foo") == 3


@pytest.mark.parametrize("detail", [False, True])
def test_ls(storage_dir: TmpDir, fs: WebdavFileSystem, detail: bool):
    """Test that ls is compliance with fsspec."""
    assert fs.ls("", detail=detail) == []

    with pytest.raises(FileNotFoundError):
        fs.ls("not-existing")

    storage_dir.gen({"empty_dir": {}})
    assert fs.ls("empty_dir", detail=detail) == []

    # try ls file
    storage_dir.gen({"foo": "foo"})
    with pytest.raises(NotADirectoryError):
        fs.ls("foo", detail=detail)

    def get_files(lst: List[Union[str, Dict[str, Any]]]) -> Set[str]:
        return {
            item["name"] if isinstance(item, dict) else item for item in lst
        }

    # try ls root
    assert get_files(fs.ls("", detail=detail)) == {"empty_dir", "foo"}

    # try ls dir with files
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    assert get_files(fs.ls("data", detail=detail)) == {"data/foo", "data/bar"}

    # try ls with files and subdirs
    storage_dir.gen(
        {"data2": {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}}
    )
    assert get_files(fs.ls("data2", detail=detail)) == {
        "data2/foo",
        "data2/bar",
        "data2/baz",
    }

    # ls with files having same name as parent
    storage_dir.gen({"lorem": {"lorem": "lorem"}})
    assert get_files(fs.ls("lorem", detail=detail)) == {"lorem/lorem"}

    # ls with subdirs having same name as parent
    storage_dir.gen({"ipsum": {"ipsum": {"ipsum": "ipsum"}}})
    assert get_files(fs.ls("ipsum", detail=detail)) == {"ipsum/ipsum"}


def test_find(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that find is compliance with fsspec."""
    storage_dir.gen(
        {
            "data": {
                "foo": "foo",
                "bar": "bar",
                "empty": {},
                "baz": {"foobaz": "foobaz"},
            }
        }
    )
    assert set(fs.find("")) == {"data/foo", "data/bar", "data/baz/foobaz"}
    assert set(fs.find("", withdirs=True)) == {
        "data",
        "data/foo",
        "data/bar",
        "data/empty",
        "data/baz",
        "data/baz/foobaz",
    }

    assert set(fs.find("", maxdepth=1)) == set()
    assert set(fs.find("", maxdepth=1, withdirs=True)) == {"data"}

    assert set(fs.find("", maxdepth=2)) == {"data/foo", "data/bar"}
    assert set(fs.find("", maxdepth=2, withdirs=True)) == {
        "data/foo",
        "data/bar",
        "data",
        "data/baz",
        "data/empty",
    }

    assert set(fs.find("not-existing")) == set()
    assert set(fs.find("data/foo")) == {"data/foo"}


def test_info(storage_dir: TmpDir, fs: WebdavFileSystem, server_address: URL):
    """Test that info is compliance with fsspec."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar", "empty": {}}})
    data_stat = (storage_dir / "data").stat()
    foo_stat = (storage_dir / "data" / "foo").stat()
    bar_stat = (storage_dir / "data" / "bar").stat()
    empty_stat = (storage_dir / "data" / "bar").stat()

    d = fs.info("data")
    assert d.pop("etag") is None
    assert d == {
        "size": None,
        "created": approx_datetime(
            datetime.fromtimestamp(int(data_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(data_stat.st_mtime), tz=timezone.utc)
        ),
        "content_language": None,
        "content_type": None,
        "type": "directory",
        "name": "data",
        "display_name": "data",
        "href": join_url(server_address, "data").path + "/",
    }

    with pytest.raises(FileNotFoundError):
        fs.info("not-existing")

    d = fs.info("data/foo")
    assert d.pop("etag")
    assert d == {
        "name": "data/foo",
        "href": join_url(server_address, "data/foo").path,
        "size": 3,
        "created": approx_datetime(
            datetime.fromtimestamp(int(foo_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(foo_stat.st_ctime), tz=timezone.utc)
        ),
        "display_name": "foo",
        "content_language": None,
        "content_type": "application/octet-stream",
        "type": "file",
    }

    d = fs.info("data/bar")
    assert d.pop("etag")
    assert d == {
        "name": "data/bar",
        "href": join_url(server_address, "data/bar").path,
        "size": 3,
        "created": approx_datetime(
            datetime.fromtimestamp(int(bar_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(bar_stat.st_ctime), tz=timezone.utc)
        ),
        "content_language": None,
        "content_type": "application/octet-stream",
        "type": "file",
        "display_name": "bar",
    }

    d = fs.info("data/empty")
    assert d.pop("etag") is None
    assert d == {
        "size": None,
        "created": approx_datetime(
            datetime.fromtimestamp(int(empty_stat.st_ctime), tz=timezone.utc)
        ),
        "modified": approx_datetime(
            datetime.fromtimestamp(int(empty_stat.st_mtime), tz=timezone.utc)
        ),
        "content_language": None,
        "content_type": None,
        "type": "directory",
        "name": "data/empty",
        "display_name": "empty",
        "href": join_url(server_address, "data/empty").path + "/",
    }


def test_walk(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that walk is compliance with fsspec."""
    assert list(fs.walk("not-existing")) == []

    storage_dir.gen({"data": {"foo": "foo"}})

    assert list(fs.walk("data/foo")) == []
    assert list(fs.walk("data")) == [
        ("data", [], ["foo"]),
    ]

    assert list(fs.walk("")) == [
        ("", ["data"], []),
        ("data", [], ["foo"]),
    ]


def test_mkdir_with_no_create_parents(
    storage_dir: TmpDir, fs: WebdavFileSystem
):
    """Test that mkdir is compliant with fsspec."""
    fs.mkdir("data", create_parents=False)
    assert storage_dir.cat() == {"data": {}}
    storage_dir.gen({"data": {"foo": "foo"}})

    with pytest.raises(FileExistsError):
        fs.mkdir("data", create_parents=False)

    with pytest.raises(NotADirectoryError):
        fs.mkdir("data/foo/bar", create_parents=False)

    with pytest.raises(FileNotFoundError):
        fs.mkdir("data/bar/bar", create_parents=False)

    assert storage_dir.cat() == {"data": {"foo": "foo"}}


def test_makedirs_exist_ok(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that makedirs with exist_ok=True is compliant with fsspec."""
    fs.makedirs("dir1/dir2/dir3/dir4", exist_ok=True)
    assert storage_dir.cat() == {"dir1": {"dir2": {"dir3": {"dir4": {}}}}}
    # should not raise any issues again
    fs.makedirs("dir1/dir2/dir3/dir4", exist_ok=True)

    fs.makedirs("dir1/dir2/dir5/dir6", exist_ok=True)
    assert storage_dir.cat() == {
        "dir1": {"dir2": {"dir3": {"dir4": {}}, "dir5": {"dir6": {}}}}
    }
    # should not raise any issues again
    fs.makedirs("dir1/dir2/dir5/dir6", exist_ok=True)

    storage_dir.gen({"data": {"foo": "foo"}})
    fs.makedirs("data", exist_ok=True)

    with pytest.raises(NotADirectoryError):
        fs.makedirs("data/foo/bar", exist_ok=True)

    with pytest.raises(FileExistsError):
        fs.makedirs("data/foo", exist_ok=True)

    assert storage_dir.cat() == {
        "data": {"foo": "foo"},
        "dir1": {"dir2": {"dir3": {"dir4": {}}, "dir5": {"dir6": {}}}},
    }


def test_makedirs_not_exist_ok(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test that makedirs with exist_ok=False is compliant with fsspec."""
    fs.makedirs("dir1/dir2/dir3/dir4", exist_ok=False)
    assert storage_dir.cat() == {"dir1": {"dir2": {"dir3": {"dir4": {}}}}}

    with pytest.raises(FileExistsError):
        fs.makedirs("dir1/dir2/dir3/dir4", exist_ok=False)

    fs.makedirs("dir1/dir2/dir5/dir6", exist_ok=False)
    assert storage_dir.cat() == {
        "dir1": {"dir2": {"dir3": {"dir4": {}}, "dir5": {"dir6": {}}}}
    }

    with pytest.raises(FileExistsError):
        fs.makedirs("dir1/dir2/dir5/dir6", exist_ok=False)

    storage_dir.gen({"data": {"foo": "foo"}})

    with pytest.raises(FileExistsError):
        fs.makedirs("data", exist_ok=False)

    with pytest.raises(NotADirectoryError):
        fs.makedirs("data/foo/bar", exist_ok=False)

    with pytest.raises(FileExistsError):
        fs.makedirs("data/foo", exist_ok=False)

    assert storage_dir.cat() == {
        "data": {"foo": "foo"},
        "dir1": {"dir2": {"dir3": {"dir4": {}}, "dir5": {"dir6": {}}}},
    }


@pytest.mark.parametrize("recursive", [True, False])
def test_copy(storage_dir: TmpDir, fs: WebdavFileSystem, recursive: bool):
    """Test copy."""
    files = {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    storage_dir.gen({"data": files})

    fs.copy("data", "data2", recursive=recursive)
    assert storage_dir.cat() == {
        "data": files,
        "data2": files if recursive else {},
    }

    fs.copy("data/foo", "data/foobar", recursive=recursive)
    assert storage_dir.cat() == {
        "data": {
            "foo": "foo",
            "foobar": "foo",
            "bar": "bar",
            "baz": {"foobaz": "foobaz"},
        },
        "data2": files if recursive else {},
    }


def test_try_copying_not_existing_file(
    storage_dir: TmpDir, fs: WebdavFileSystem
):
    """Test trying to copy a not existing file raises FileNotFoundError."""
    with pytest.raises(FileNotFoundError):
        fs.copy("not-existing-file", "bar")

    with pytest.raises(FileNotFoundError):
        fs.cp_file("not-existing-file", "bar")

    assert storage_dir.cat() == {}


@pytest.mark.parametrize("recursive", [True, False])
def test_remove_dir(
    storage_dir: TmpDir, fs: WebdavFileSystem, recursive: bool
):
    """Test rm directory."""
    # this is where we deviate a bit from LocalFS
    # it seems we need to override a lot of things to achieve this,
    # as `fs.rm` uses expand_paths that also returns dirs, and is passed
    # to `fs.rm_file` which is not supposed to remove a directory.
    files = {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    storage_dir.gen({"data": files})

    fs.rm("data/baz", recursive=recursive)
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}


@pytest.mark.parametrize("recursive", [True, False])
def test_remove_file(
    storage_dir: TmpDir, fs: WebdavFileSystem, recursive: bool
):
    """Test rm file."""
    files = {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    storage_dir.gen({"data": files})

    fs.rm("data/foo", recursive=recursive)

    expected = {k: v for k, v in files.items() if k != "foo"}
    assert storage_dir.cat() == {"data": expected}


def test_try_removing_not_existing_file(
    storage_dir: TmpDir, fs: WebdavFileSystem
):
    """Test rm and rm_file on non-existing file raises FileNotFoundError."""
    with pytest.raises(FileNotFoundError):
        fs.rm("not-existing-file")

    with pytest.raises(FileNotFoundError):
        fs.rm_file("not-existing-file")

    assert storage_dir.cat() == {}


def test_rmdir(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test rmdir."""
    storage_dir.gen({"data": {"foo": "foo"}})

    with pytest.raises(OSError) as exc_info:
        fs.rmdir("data")

    assert exc_info.value.errno == errno.ENOTEMPTY
    assert exc_info.value.filename == "data"

    with pytest.raises(NotADirectoryError) as exc_info:
        fs.rmdir("data/foo")

    assert exc_info.value.errno == errno.ENOTDIR
    assert exc_info.value.filename == "data/foo"


def test_rmdir_empty(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test rmdir success on empty dir."""
    storage_dir.gen({"data": {}})

    fs.rmdir("data")
    assert storage_dir.cat() == {}


def test_rmdir_not_existing(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test rmdir on non-existing file raises FileNotFoundError."""
    with pytest.raises(FileNotFoundError):
        fs.rmdir("not-existing-dir/file")
    assert storage_dir.cat() == {}


def test_move_no_recursive(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test move without recursive."""
    files = {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    storage_dir.gen({"data": files})

    fs.mv("data", "data2", recursive=False)
    assert storage_dir.cat() == {
        "data2": {},
        "data": files,
    }

    fs.mv("data/foo", "data2/foobar", recursive=False)
    assert storage_dir.cat() == {
        "data2": {"foobar": "foo"},
        "data": {"bar": "bar", "baz": {"foobaz": "foobaz"}},
    }


def test_move_recursive(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test move recursive."""
    files = {"foo": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    storage_dir.gen({"data": files})

    fs.mv("data", "data2", recursive=True)
    assert storage_dir.cat() == {"data2": files}

    fs.mv("data2/foo", "data2/foobar", recursive=True)
    assert storage_dir.cat() == {
        "data2": {"foobar": "foo", "bar": "bar", "baz": {"foobaz": "foobaz"}}
    }


def test_try_moving_not_existing_file(
    storage_dir: TmpDir, fs: WebdavFileSystem
):
    """Test trying to move a non-existing file raises FileNotFoundError."""
    with pytest.raises(FileNotFoundError):
        fs.mv("not-existing-file", "bar")

    assert storage_dir.cat() == {}


def test_touch_truncates(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test touch with truncate=True."""
    fs.touch("foo")
    assert storage_dir.cat() == {"foo": ""}

    storage_dir.gen({"foo": "foo"})
    fs.touch("foo")
    assert storage_dir.cat() == {"foo": ""}


def test_touch_not_truncate(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test touch not truncate."""
    fs.touch("foo", truncate=False)
    assert storage_dir.cat() == {"foo": ""}

    with pytest.raises(NotImplementedError):
        # this should update the timestamp, but it's not implemented yet.
        fs.touch("foo", truncate=False)


@pytest.mark.parametrize("with_size", [True, False])
@pytest.mark.parametrize("wrap_fobj", [True, False])
def test_upload_fileobj(
    storage_dir: TmpDir,
    fs: WebdavFileSystem,
    with_size: bool,
    wrap_fobj: bool,
):
    """Test upload_fileobj.

    If with_size, size hints are provided to the API.
    If wrap_fobj, the file object is wrapped with the ReadWrapper
       that makes the peek_fileobj_length util to not be able to
       figure out the size of the given fileobj. For the most part,
       it should still work in this case.
    """
    if wrap_fobj and not with_size:
        pytest.skip("the test server does not work without content-length :(")

    foo = storage_dir / "foo"
    length = foo.write_bytes(b"foo")

    with foo.open(mode="rb") as fobj:
        if wrap_fobj:
            fobj = ReadWrapper(fobj)  # type: ignore
        size = length if with_size else None
        fs.upload_fileobj(fobj, "data/foobar", size=size)

    assert (storage_dir / "data").cat() == {"foobar": "foo"}


def test_callbacks(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test fsspec callbacks."""
    src_file = storage_dir / "source"
    dest_file = "data/get_put_file/dest"

    data = b"test" * 4
    size = src_file.write_bytes(data)

    class EventLogger(fsspec.Callback):
        """Log callback values."""

        def __init__(self) -> None:
            self.events: List[Tuple[str, int]] = []
            super().__init__()

        def set_size(self, size: int) -> None:
            """Log a set_size call."""
            self.events.append(("set_size", size))

        def relative_update(self, value: int) -> None:
            """Log a relative_update call."""
            self.events.append(("relative_update", value))

    event_logger = EventLogger()
    fs.put_file(src_file, dest_file, chunk_size=4, callback=event_logger)
    assert (storage_dir / dest_file).cat() == data.decode()

    assert event_logger.events[0] == ("set_size", size)
    assert event_logger.events[1:] == [
        ("relative_update", 4),
        ("relative_update", 4),
        ("relative_update", 4),
        ("relative_update", 4),
        ("relative_update", 0),
    ]



"""Utilities to wrap file objects for callback purposes."""
from collections.abc import Iterable
from functools import wraps
from typing import IO, TYPE_CHECKING, Any, AnyStr, Callable, Iterator, cast

if TYPE_CHECKING:
    from typing_extensions import Literal


CallbackFn = Callable[[int], Any]


def do_nothing(_: int) -> Any:
    """A function that does nothing."""


class CallbackIOWrapper(Iterable):
    """Wrap a file-like's read/write method to report length to callback."""

    def __init__(
        self,
        stream: IO[AnyStr],
        callback: CallbackFn = None,
        method: "Literal['read', 'write']" = "read",
    ) -> None:
        """Pass stream and callback and appropriate method to wrap."""
        if method not in {"read", "write"}:
            raise ValueError("Can only wrap read/write methods")

        cb_func = callback or do_nothing
        assert hasattr(stream, method)

        if method == "read":
            sread = stream.read

            @wraps(sread)
            def fread(*args: Any, **kw: Any) -> AnyStr:
                result = sread(*args, **kw)
                cb_func(len(result))
                return result

            self.__dict__[method] = fread

        else:
            swrite = stream.write

            @wraps(swrite)
            def fwrite(chunk: AnyStr, *args: Any, **kw: Any) -> int:
                result = swrite(chunk, *args, **kw)  # type: ignore[call-arg]
                cb_func(len(chunk))
                return result

            self.__dict__[method] = fwrite

        self.__dict__.update(
            {
                "__wrapped_stream__": stream,
                "__wrapped_method__": method,
                "__call_back__": cb_func,
            }
        )

    def __iter__(self) -> Iterator[AnyStr]:
        """Iterate through the stream and call callback method with size.

        Note: We need this because, HTTPX uses `in iter` when uploading files.
        """
        stream = self.__wrapped_stream__
        method = self.__wrapped_method__
        try:
            chunks = iter(stream)
        except TypeError:
            if method != "read":
                raise

            # pylint: disable=import-outside-toplevel
            from .stream import read_until

            chunks = read_until(stream, "\n")

        for chunk in chunks:
            self.__call_back__(len(chunk))
            yield chunk

    def __setattr__(self, attr: str, value: Any) -> Any:
        """Setting attr on the stream."""
        setattr(self.__wrapped_stream__, attr, value)

    def __getattr__(self, attr: str) -> Any:
        """Getting attr on the stream."""
        return getattr(self.__wrapped_stream__, attr)


def wrap_file_like(
    file_obj: IO[AnyStr],
    callback: CallbackFn = None,
    method: "Literal['read', 'write']" = "read",
) -> IO[AnyStr]:
    """Wrap a file-like object that reports to callback on r/w operation.

    This method exists to cast the type to the file-like on return.
    """
    wrapper = CallbackIOWrapper(file_obj, callback, method=method)
    return cast(IO[AnyStr], wrapper)



"""Testing common functional utilities."""

import inspect
from unittest import mock

import pytest

from webdav4.func_utils import repeat_func, retry, wrap_fn


def test_repeat_func():
    """Test repeat_func that it gets repeatedly called over n times(or inf)."""

    def always_true():
        return True

    assert list(repeat_func(always_true, times=10)) == [True] * 10
    # infinite generator
    it = repeat_func(always_true)
    assert next(it)
    assert next(it)


def test_wrap_fn():
    """Test wrap_fn which reduces arity to zero."""
    func = wrap_fn(sum, [1, 2, 3])

    params = inspect.getfullargspec(func)
    assert not any(params)
    assert func() == 6


def test_retry():
    """Test retry."""
    value = 3

    func = mock.MagicMock(side_effect=[KeyError, IndexError, value])
    wrapped = retry(3, (KeyError, IndexError), timeout=0.01)

    assert wrapped(func) == value
    assert func.call_count == 3


def test_retry_all_attempts_fail():
    """Test retry but all attempts fail."""
    value = 3

    func = mock.MagicMock(side_effect=[KeyError, IndexError, ValueError])
    wrapped = retry(3, (KeyError, IndexError, ValueError), timeout=0.01)

    with pytest.raises(ValueError):
        assert wrapped(func) == value
    assert func.call_count == 3


def test_retry_unhandled_errors():
    """Test retry for unhandled errors."""
    value = 3

    func = mock.MagicMock(side_effect=[KeyError, IndexError])
    wrapped = retry(3, [KeyError], timeout=0.01)

    with pytest.raises(IndexError):
        assert wrapped(func) == value
    assert func.call_count == 2


def test_retry_once():
    """Test retry once, i.e. no retry at all."""
    value = 3

    func = mock.MagicMock(side_effect=[KeyError])
    wrapped = retry(1, [KeyError], timeout=0.01)

    with pytest.raises(KeyError):
        assert wrapped(func) == value
    assert func.call_count == 1


def test_retry_filter_errors():
    """Test retry with extra filter for errors."""
    value = 3
    _filter = mock.MagicMock(return_value=True)

    excs = [KeyError, IndexError, ValueError]
    func = mock.MagicMock(side_effect=excs)
    wrapped = retry(3, excs, timeout=0.01, filter_errors=_filter)

    with pytest.raises(ValueError):
        assert wrapped(func) == value

    assert all(
        isinstance(val[0][0], typ)
        for val, typ in zip(_filter.call_args_list, excs)
    )

    assert func.call_count == 3


def test_retry_filter_errors_raises_false():
    """Test retry with extra filter that returns false (asking to raise)."""
    value = 3
    _filter = mock.MagicMock(return_value=False)

    excs = [KeyError, IndexError, ValueError]
    func = mock.MagicMock(side_effect=excs)
    wrapped = retry(3, excs, timeout=0.01, filter_errors=_filter)

    with pytest.raises(KeyError):
        assert wrapped(func) == value

    assert func.call_count == 1



"""URLs parsing logics here."""
from re import sub

from httpx import URL


def strip_leading_slash(path: str) -> str:
    """Strips leading slash from the path, except when it's a root."""
    return path.rstrip("/") if path and path != "/" else path


def normalize_path(path: str) -> str:
    """Normalizes path, removes leading slash."""
    path = sub("/{2,}", "/", path)
    return strip_leading_slash(path)


def join_url(
    base_url: URL, path: str, add_trailing_slash: bool = False
) -> URL:
    """Joins base url with a path."""
    base_path = base_url.path
    path = join_url_path(base_path, path)
    if add_trailing_slash:
        path += "/"
    return base_url.copy_with(path=path)


def join_url_path(hostname: str, path: str) -> str:
    """Returns path absolute."""
    path = path.strip("/")
    return normalize_path(f"/{hostname}/{path}")


def relative_url_to(base_url: URL, rel: str) -> str:
    """Finds relative url to a base url path."""
    base = base_url.path.strip("/")
    rel = rel.strip("/")

    if base == rel or not rel:
        return "/"

    if not base and rel:
        return rel

    index = len(base) + 1
    return rel[index:]



"""fsspec compliant webdav file system."""
import errno
import io
import os
import tempfile
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterator,
    List,
    NamedTuple,
    NoReturn,
    Optional,
    TextIO,
    Tuple,
    Type,
    Union,
    cast,
)

from fsspec import Callback
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem

from .client import (
    Client,
    IsACollectionError,
    IsAResourceError,
    ResourceAlreadyExists,
    ResourceConflict,
    ResourceNotFound,
)
from .fs_utils import peek_filelike_length
from .stream import read_into

if TYPE_CHECKING:
    from array import ArrayType
    from datetime import datetime
    from mmap import mmap
    from os import PathLike
    from typing import AnyStr

    from .callback import CallbackFn
    from .types import AuthTypes, URLTypes


mapping = {"content_length": "size", "path": "name", "type": "type"}


def translate_info(item: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
    """Translate info from the client to as per fsspec requirements."""
    assert not isinstance(item, str)
    return {mapping.get(key, key): value for key, value in item.items()}


@contextmanager
def translate_exceptions() -> Iterator[None]:
    """Translate exceptions from Client to the fsspec compatible one."""
    try:
        yield
    except ResourceNotFound as exc:
        raise FileNotFoundError(
            errno.ENOENT, "No such file or directory", exc.path
        ) from exc
    except IsACollectionError as exc:
        raise IsADirectoryError(
            errno.EISDIR, "Is a directory", exc.path
        ) from exc
    except IsAResourceError as exc:
        raise NotADirectoryError(
            errno.ENOTDIR, "Not a directory", exc.path
        ) from exc


class WebdavFileSystem(AbstractFileSystem):
    """Provides access to webdav through fsspec-compliant APIs."""

    protocol = ("webdav", "dav")

    def __init__(
        self,
        base_url: "URLTypes",
        auth: "AuthTypes" = None,
        client: "Client" = None,
        **client_opts: Any,
    ) -> None:
        """Instantiate WebdavFileSystem with base_url and auth.

        Args:
            base_url: base url of the server
            auth: Authentication to the server
                Refer to HTTPX's auth for more information.
            client: Webdav client to use instead, useful for testing/mocking,
                or extending WebdavFileSystem.
            client_opts: Extra args that are passed to Webdav Client.
                (refer to it's documenting for more information).
        """
        super().__init__()
        client_opts.setdefault("chunk_size", self.blocksize)
        self.client = client or Client(base_url, auth=auth, **client_opts)

    @classmethod
    def _strip_protocol(cls, path: str) -> str:
        """Strips protocol from the given path, overriding for type-casting."""
        stripped = super()._strip_protocol(path)
        return cast(str, stripped)

    @translate_exceptions()
    def ls(
        self, path: str, detail: bool = True, **kwargs: Any
    ) -> List[Union[str, Dict[str, Any]]]:
        """`ls` implementation for fsspec, see fsspec for more information."""
        path = self._strip_protocol(path).strip()
        data = self.client.ls(
            path, detail=detail, allow_listing_resource=False
        )
        if not detail:
            return data
        return [translate_info(item) for item in data]

    @translate_exceptions()
    def info(self, path: str, **kwargs: Any) -> Dict[str, Any]:
        """Return information about the current path."""
        path = self._strip_protocol(path)
        return translate_info(self.client.info(path))

    @translate_exceptions()
    def rm_file(self, path: str) -> None:
        """Remove a file."""
        path = self._strip_protocol(path)
        # not checking if it's a directory as `rm` also passes a directory
        return self.client.remove(path)

    _rm = rm_file

    @translate_exceptions()
    def cp_file(self, path1: str, path2: str, **kwargs: Any) -> None:
        """Copy a file from one path to the other."""
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)
        # not checking if it's a directory as `cp` also passes a directory
        return self.client.copy(path1, path2)

    def rmdir(self, path: str) -> None:
        """Remove a directory, if empty."""
        path = self._strip_protocol(path)
        if self.ls(path):
            raise OSError(errno.ENOTEMPTY, "Directory not empty", path)
        return self.client.remove(path)

    def rm(
        self, path: str, recursive: bool = False, maxdepth: int = None
    ) -> None:
        """Delete files and directories."""
        path = self._strip_protocol(path)
        if recursive and not maxdepth and self.isdir(path):
            return self.rm_file(path)
        super().rm(path, recursive=recursive, maxdepth=maxdepth)
        return None

    def copy(
        self,
        path1: str,
        path2: str,
        recursive: bool = False,
        on_error: str = None,
        **kwargs: Any,
    ) -> None:
        """Copy files and directories."""
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)

        if recursive and not kwargs.get("maxdepth") and self.isdir(path1):
            return self.cp_file(path1, path2)

        if not recursive and self.isdir(path1):
            return self.makedirs(path2)

        super().copy(
            path1, path2, recursive=recursive, on_error=on_error, **kwargs
        )
        return None

    def mv(
        self,
        path1: str,
        path2: str,
        recursive: bool = False,
        maxdepth: bool = None,
        **kwargs: Any,
    ) -> None:
        """Move a file/directory from one path to the other."""
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)

        if recursive and not maxdepth and self.isdir(path1):
            return self.client.move(path1, path2)

        if not recursive and self.isdir(path1):
            return self.makedirs(path2)

        super().mv(
            path1, path2, recursive=recursive, maxdepth=maxdepth, **kwargs
        )
        return None

    def _mkdir(self, path: str, exist_ok: bool = False) -> None:
        """Creates directory and translates to an appropriate exceptions.

        Internally, Client tries to do as much less API call as possible.
        And, usually, the ResourceAlreadyExists and ResourceConflict are
        enough for a standard server. But, in fsspec, for better exceptions,
        we have to distinguish between different conditions:

        1. parent not being a directory (NotADirectoryError),
        2. parent does not exists (FileExistsError),
        3. path already exists (FileNotFoundError), etc.

        We also spend some API calls to be sure on error.
        """
        try:
            return self.client.mkdir(path)
        except ResourceAlreadyExists as exc:
            details = self.info(path)
            if details and details["type"] == "directory" and exist_ok:
                return None

            raise FileExistsError(errno.EEXIST, "File exists", path) from exc
        except ResourceConflict as exc:
            parent = self._parent(path)
            details = self.info(parent)
            if details["type"] == "directory":
                raise  # pragma: no cover
            raise NotADirectoryError(
                errno.ENOTDIR, "Not a directory", parent
            ) from exc

    def mkdir(
        self, path: str, create_parents: bool = True, **kwargs: Any
    ) -> None:
        """Create directory."""
        path = self._strip_protocol(path)
        if create_parents:
            return self.makedirs(path, exist_ok=True)
        return self._mkdir(path)

    def makedirs(self, path: str, exist_ok: bool = False) -> None:
        """Creates directory to the given path."""
        path = self._strip_protocol(path)
        parent = self._parent(path)
        if not ({"", self.root_marker} & {path, parent}) and not self.exists(
            parent
        ):
            self.makedirs(parent, exist_ok=exist_ok)

        return self._mkdir(path, exist_ok=exist_ok)

    @translate_exceptions()
    def created(self, path: str) -> Optional["datetime"]:
        """Returns creation time/date."""
        path = self._strip_protocol(path)
        return self.client.created(path)

    @translate_exceptions()
    def modified(self, path: str) -> Optional["datetime"]:
        """Returns last modified time/data."""
        path = self._strip_protocol(path)
        return self.client.modified(path)

    @translate_exceptions()
    def _open(
        self,
        path: str,
        mode: str = "rb",
        block_size: int = None,
        autocommit: bool = True,
        cache_options: Dict[str, str] = None,
        **kwargs: Any,
    ) -> Union["WebdavFile", "UploadFile"]:
        """Return a file-like object from the filesystem."""
        size = kwargs.pop("size", None)
        assert "a" not in mode

        if "x" in mode and self.exists(path):
            raise FileExistsError(errno.EEXIST, "File exists", path)
        if set(mode) & {"w", "x"}:
            return UploadFile(
                self, path=path, mode=mode, block_size=block_size
            )

        return WebdavFile(
            self,
            path,
            block_size=block_size,
            autocommit=autocommit,
            mode=mode,
            size=size,
            cache_options=cache_options,
            **kwargs,
        )

    @translate_exceptions()
    def checksum(self, path: str) -> Optional[str]:
        """Returns checksum/etag of the path."""
        path = self._strip_protocol(path)
        return self.client.etag(path)

    @translate_exceptions()
    def size(self, path: str) -> Optional[int]:
        """Returns size of the path."""
        path = self._strip_protocol(path)
        return self.client.content_length(path)

    def sign(self, path: str, expiration: int = 100, **kwargs: Any) -> None:
        """Create a signed URL representing the given path."""
        raise NotImplementedError

    def pipe_file(self, path: str, value: bytes, **kwargs: Any) -> None:
        """Upload the contents to given file in the remote webdav server."""
        buff = io.BytesIO(value)
        kwargs.setdefault("overwrite", True)
        return self.upload_fileobj(buff, path, **kwargs)

    def upload_fileobj(
        self,
        fobj: BinaryIO,
        rpath: str,
        callback: "Callback" = None,
        overwrite: bool = True,
        size: int = None,
        **kwargs: Any,
    ) -> None:
        """Upload contents from the fileobj to the remote path."""
        rpath = self._strip_protocol(rpath)
        self.mkdirs(os.path.dirname(rpath), exist_ok=True)

        if size is None:
            size = peek_filelike_length(fobj)

        callback = cast("Callback", Callback.as_callback(callback))
        if size is not None:  # pragma: no cover
            callback.set_size(size)
        progress_callback = cast("CallbackFn", callback.relative_update)

        return self.client.upload_fileobj(
            fobj,
            rpath,
            overwrite=overwrite,
            callback=progress_callback,
            size=size,
            **kwargs,
        )

    put_fileobj = upload_fileobj

    def put_file(
        self,
        lpath: "PathLike[AnyStr]",
        rpath: str,
        callback: "Callback" = None,
        **kwargs: Any,
    ) -> None:
        """Copy file to remote webdav server."""
        if os.path.isdir(lpath):
            rpath = self._strip_protocol(rpath)
            return self.makedirs(rpath, exist_ok=True)

        with open(lpath, mode="rb") as fobj:
            kwargs.setdefault("overwrite", True)
            kwargs.setdefault("size", None)
            return self.upload_fileobj(
                fobj,
                rpath,
                callback=callback,
                **kwargs,
            )


class WebdavFile(AbstractBufferedFile):
    """WebdavFile that provides file-like access to remote file."""

    size: int

    def __init__(
        self,
        fs: "WebdavFileSystem",
        path: str,
        mode: str = "rb",
        block_size: int = None,
        autocommit: bool = True,
        cache_type: str = "readahead",
        cache_options: Dict[str, str] = None,
        **kwargs: Any,
    ) -> None:
        """Instantiate a file-like object with the provided options.

        See fsspec for more information.
        """
        size = kwargs.get("size")
        self.details = {"name": path, "size": size, "type": "file"}
        super().__init__(
            fs,
            path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            **kwargs,
        )
        encoding = kwargs.get("encoding")
        self.fobj = fs.client.open(
            self.path,
            mode=self.mode,
            encoding=encoding,
            chunk_size=self.blocksize,
        )
        self.reader: Union[TextIO, BinaryIO] = self.fobj.__enter__()

        # only get the file size if GET request didnot send Content-Length
        # or was retrieved before.
        if not self.size:
            if getattr(self.reader, "size", None):
                self.size = self.reader.size  # type: ignore
            else:
                self.size = self.fs.size(self.path)

        self.closed: bool = False

    def read(self, length: int = -1) -> Union[str, bytes, None]:
        """Read chunk of bytes."""
        chunk = self.reader.read(length)
        if chunk:
            self.loc += len(chunk)
        return chunk

    def __enter__(self) -> "WebdavFile":
        """Start streaming."""
        return self

    def _fetch_range(self, start: int, end: int) -> None:
        """Not essential. Creating stub to make pylint happy."""
        raise NotImplementedError

    def seek(self, loc: int, whence: int = 0) -> int:
        """Set current file location."""
        super().seek(loc, whence=whence)
        return self.reader.seek(loc, whence=whence)

    def isatty(self) -> bool:
        """Check if it is an interactive fileobj."""
        return False

    def close(self) -> None:
        """Close stream."""
        if self.closed:
            return
        if hasattr(self, "reader"):
            # fs.client.open might have raised an error
            self.reader.close()
        self.closed = True

    def __reduce_ex__(
        self, protocol: int
    ) -> Tuple[Callable[["ReopenArgs"], Type["WebdavFile"]], "ReopenArgs"]:
        """Recreate/reopen file when restored."""
        return reopen, ReopenArgs(  # pragma: no cover
            WebdavFile,
            self.fs,
            self.path,
            self.blocksize,
            self.mode,
            self.size,
        )


class ReopenArgs(NamedTuple):  # pylint: disable=inherit-non-class
    """Args to reopen the file."""

    file: Type[WebdavFile]
    fs: "WebdavFileSystem"
    path: str
    blocksize: Optional[int]
    mode: str
    size: Optional[int]


def reopen(args: ReopenArgs) -> WebdavFile:
    """Reopen file when unpickled."""
    return args.file(  # pragma: no cover
        args.fs,
        args.path,
        blocksize=args.blocksize,
        mode=args.mode,
        size=args.size,
    )


# TODO: if we need this to be serializable, we might want to use tempfile
#  directly.


class UploadFile(tempfile.SpooledTemporaryFile):
    """UploadFile for Webdav, that uses SpooledTemporaryFile.

    In Webdav, you cannot upload in chunks. Similar to http, we need some kind
    of protocol on top to be able to do that. Sabredav provides a way to upload
    chunks, but it's not implemented in Owncloud and Nextcloud. They have a
    different chunking mechanism. On top of it, there are TUS implementations
    in newer version of Owncloud.

    However, as they are only for chunking, it needs to be seen if it is
    possible to implement full-breadth of FileObj API through that (I doubt it)

    Note that, fsspec's put_file/pipe_file don't use this technique and are
    thus faster as they don't need buffering. It is recommended to use that
    if you don't need to upload content dynamically.
    """

    def __init__(  # pylint: disable=invalid-name
        self,
        fs: "WebdavFileSystem",
        path: str,
        mode: str = "wb",
        block_size: int = None,
    ):
        """Extended interface with path and fs."""
        assert fs
        assert path
        self.blocksize = (
            AbstractBufferedFile.DEFAULT_BLOCK_SIZE
            if block_size in ["default", None]
            else block_size
        )
        self.fs: WebdavFileSystem = fs  # pylint: disable=invalid-name
        assert mode
        self.path: str = path

        # whatever the mode be, we should try to open the file
        # in both rw mode.
        super().__init__(max_size=self.blocksize, mode="wb+")

    def __exit__(self, *exc: Any) -> None:
        """Upload file by seeking to first byte on exit."""
        self.close()

    def readable(self) -> bool:  # pylint: disable=no-self-use
        """It is readable."""
        return True

    def writable(self) -> bool:  # pylint: disable=no-self-use
        """It is writable."""
        return True

    def seekable(self) -> bool:  # pylint: disable=no-self-use
        """It is seekable."""
        return True

    def commit(self) -> None:
        """Commits the file to the given path.

        As we cannot upload in chunk, this is where we really upload file.
        """
        self.seek(0)
        fileobj = cast(BinaryIO, self)
        self.fs.client.upload_fileobj(
            fileobj, self.path, chunk_size=self.blocksize, overwrite=True
        )

    def close(self) -> None:
        """Close the file."""
        if not self.closed:
            self.commit()
            super().close()

    def discard(self) -> None:
        """Discard the file."""
        if not self.closed:
            super().close()

    def info(self) -> NoReturn:  # pylint: disable=no-self-use
        """Info about the file upload that is in progress."""
        raise ValueError("cannot provide info in write-mode")

    def readinto(
        self, sequence: Union[bytearray, memoryview, "ArrayType[Any]", "mmap"]
    ) -> int:
        """Read bytes into the given buffer."""
        return read_into(self, sequence)

    def readuntil(self, char: bytes = b"\n", blocks: int = None) -> bytes:
        """Read until the given character is found."""
        ret = AbstractBufferedFile.readuntil(self, char=char, blocks=blocks)
        return cast(bytes, ret)



"""Testing stream utilities."""
from io import DEFAULT_BUFFER_SIZE, BytesIO, StringIO
from typing import Any, Iterator

import pytest
from pytest import MonkeyPatch

from tests.utils import TmpDir
from webdav4.client import Client
from webdav4.fsspec import WebdavFileSystem
from webdav4.stream import read_until


def test_read_until_non_binary():
    """Test read_until with non-binary obj."""
    buff = StringIO()
    buff.write("hello world\n" * 100)
    buff.seek(0)

    assert list(read_until(buff, "\n")) == ["hello world\n"] * 100


def test_read_until_binary():
    """Test read_until with binary obj."""
    buff = BytesIO()
    buff.write(b"hello world\n" * 100)
    buff.seek(0)

    assert list(read_until(buff, "\n")) == [b"hello world\n"] * 100


@pytest.mark.parametrize("buff", [StringIO(), BytesIO()])
def test_read_until_empty(buff):
    """Test read_until with empty data."""
    assert list(read_until(buff, "\n")) == []


def test_read_until_not_found_any_match():
    """Test read_until, with no matching character."""
    d = "hello world" * 100
    buff = StringIO(d)

    assert list(read_until(buff, "\n")) == [d]


def test_retry_reconnect_on_failure(
    storage_dir: TmpDir,
    fs: WebdavFileSystem,
    client: Client,
    monkeypatch: MonkeyPatch,
):
    """Test retry/reconnect on network failures."""
    from unittest import mock

    from webdav4.http import HTTPNetworkError, HTTPResponse

    original_iter_content = HTTPResponse.iter_bytes

    def bad_iter_content(
        response: "HTTPResponse", *args: Any, **kwargs: Any
    ) -> Iterator[bytes]:
        """Simulate bad connection."""
        it = original_iter_content(response, *args, **kwargs)
        for i, chunk in enumerate(it):
            # Drop connection error on second chunk if there is one
            if i > 0:
                raise HTTPNetworkError(
                    "Simulated connection drop", request=response.request
                )
            yield chunk

    # Text should be longer than default chunk to test resume,
    # using twice of that plus something tests second resume,
    # this is important because second response is different
    text1 = "0123456789" * (client.chunk_size // 10 + 1)
    storage_dir.gen("sample.txt", text1 * 2)
    propfind_resp = client.propfind("sample.txt")

    monkeypatch.setattr(HTTPResponse, "iter_bytes", bad_iter_content)

    with mock.patch.object(client, "propfind", return_value=propfind_resp):
        with client.open("sample.txt") as fd:
            # Test various .read() variants
            assert fd.read(len(text1)) == text1
            assert fd.read() == text1
            assert fd.read() == ""

        with client.open("sample.txt", mode="rb") as fd:
            # Test various .read() variants
            assert fd.read(len(text1)) == text1.encode()
            assert fd.read() == text1.encode()
            assert fd.read() == b""

        with fs.open("sample.txt", mode="r") as fd:
            # Test various .read() variants
            assert fd.read(len(text1)) == text1
            assert fd.read() == text1
            assert fd.read() == ""

        with fs.open("sample.txt") as fd:
            # Test various .read() variants
            assert fd.read(len(text1)) == text1.encode()
            assert fd.read() == text1.encode()
            assert fd.read() == b""

        # when we cannot detect support for ranges, we should just raise error
        client.detected_features.supports_ranges = False
        with client.open("sample.txt", mode="rb") as fd:
            fd._initial_response.headers.clear()  # type: ignore
            with pytest.raises(HTTPNetworkError):
                fd.read()

            assert fd.supports_ranges is False  # type: ignore
            with pytest.raises(ValueError) as exc_info:
                fd.seek(10)
            assert str(exc_info.value) == "server does not support ranges"

        with fs.open("sample.txt", mode="rb") as fd:
            fd.reader._initial_response.headers.clear()  # type: ignore
            with pytest.raises(HTTPNetworkError):
                fd.read()

            assert fd.reader.supports_ranges is False  # type: ignore
            with pytest.raises(ValueError) as exc_info:
                fd.seek(10)
            assert str(exc_info.value) == "server does not support ranges"


def test_open(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test opening a remote file from webdav using fs in text mode."""
    text1 = "0123456789" * (DEFAULT_BUFFER_SIZE // 10 + 1)
    storage_dir.gen("sample.txt", text1 * 2)

    with fs.open("sample.txt", mode="r") as f:
        assert not f.closed
        assert not f.isatty()
        assert f.readable()
        assert not f.writable()
        assert f.seekable()
        assert f.tell() == 0
        assert f.read(len(text1)) == text1
        assert f.tell() == len(text1)
        assert f.seek(10) == 10
        assert f.read(len(text1) - 10) == text1[10:]
        assert f.tell() == len(text1)
        assert f.seek(len(text1) * 2) == len(text1) * 2
        assert f.read() == ""
        assert f.seek(0) == 0
    assert f.closed
    f.close()

    text2 = "0123456789\n"
    storage_dir.gen("sample2.txt", text2 * 10)

    with fs.open("sample2.txt", mode="r") as f:
        assert f.readline() == text2
        assert f.readlines() == [text2] * 9
    assert f.closed
    f.close()


def test_open_binary(storage_dir: TmpDir, fs: WebdavFileSystem):
    """Test file object in binary mode with fs."""
    text1 = b"0123456789" * (DEFAULT_BUFFER_SIZE // 10 + 1)
    storage_dir.gen("sample.txt", text1 * 2)
    with fs.open("sample.txt", mode="rb") as f:
        assert not f.closed
        assert not f.isatty()
        assert f.readable()
        assert not f.writable()
        assert f.seekable()
        assert f.tell() == 0

        assert f.read(len(text1)) == text1
        assert f.tell() == len(text1)

        assert f.seek(10) == 10
        assert f.read(len(text1) - 10) == text1[10:]
        assert f.tell() == len(text1)

        assert f.seek(len(text1), 1) == len(text1) * 2
        assert f.read() == b""
        assert f.seek(-len(text1), 1) == len(text1)

        assert f.seek(0) == 0
        buff = bytearray(5)
        assert f.readinto(buff) == 5
        assert bytes(buff) == text1[:5]
        length = f.readinto1(buff)
        assert bytes(buff) == text1[5 : 5 + length]
        assert f.seek(-10, 2) == f.tell() == len(text1) * 2 - 10
        assert f.read() == text1[-10:]
    assert f.closed
    f.close()

    with fs.open("sample.txt", mode="rb") as f:
        with pytest.raises(ValueError):
            f.seek(-10)
        with pytest.raises(ValueError):
            f.seek(10, 3)
    assert f.closed
    f.close()

    text2 = b"0123456789\n"
    storage_dir.gen("sample2.txt", text2 * 10)

    with fs.open("sample2.txt", mode="rb") as f:
        assert f.readline() == text2
        assert f.readlines() == [text2] * 9
    assert f.closed
    f.close()


def test_close_connection_if_nothing_is_read(client: Client):
    """Test that http connection is closed when nothing is read."""
    from unittest.mock import MagicMock, patch

    response = MagicMock()

    with patch.object(
        client.http, "send", return_value=response
    ), patch.object(client, "isdir", return_value=False):
        with client.open("sample.txt", mode="rb"):
            response.close.assert_not_called()
        response.close.assert_called_once()



"""Tests for webdav client."""
from datetime import datetime, timezone
from http import HTTPStatus
from io import DEFAULT_BUFFER_SIZE, BytesIO
from pathlib import Path
from typing import Any, Callable, Dict
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest

from webdav4.client import (
    BadGatewayError,
    Client,
    FeatureDetection,
    ForbiddenOperation,
    HTTPError,
    InsufficientStorage,
    IsACollectionError,
    MultiStatusError,
    ResourceAlreadyExists,
    ResourceConflict,
    ResourceLocked,
    ResourceNotFound,
)
from webdav4.http import Client as HTTPClient
from webdav4.http import Method
from webdav4.http import Method as HTTPMethod
from webdav4.urls import URL

from .utils import TmpDir, approx_datetime


def lock_resource(client: Client, path: str):
    """Exclusive lock on a resource."""
    url = client.join_url(path)
    d = f"""<?xml version="1.0" encoding="utf-8" ?>
     <d:lockinfo xmlns:d='DAV:'>
       <d:lockscope><d:exclusive/></d:lockscope>
       <d:locktype><d:write/></d:locktype>
       <d:owner><d:href>{url}</d:href></d:owner>
     </d:lockinfo>"""
    resp = client.http.lock(url, content=d)
    resp.raise_for_status()


def test_init():
    """Try default initialization."""
    base_url = "http://example.org"
    auth = ("user", "password")

    c = Client(base_url=base_url, auth=auth)

    assert c.base_url, c.http.auth == (base_url, auth)


def test_init_pass_client():
    """Test passing a pre-build http client to a webdav client.

    If we test this, then it might be easier to mock in other tests.
    """
    base_url = "http://example.org/"
    auth = ("user", "password")

    http_client = HTTPClient(auth=auth)
    client = Client(base_url=base_url, http_client=http_client)

    assert client.http == http_client
    assert client.base_url == base_url


def test_get_property(storage_dir: TmpDir, client: Client):
    """Test getting a property from a resource."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    assert client.get_property("/data/foo", "content_length") == 3
    assert client.get_property("/data/", "content_length") is None
    assert client.get_property("/data/", "whatever") == ""


def test_copy_file(storage_dir: TmpDir, client: Client):
    """Test simple file copy."""
    storage_dir.gen({"data": {"foo": b"foo"}})
    client.copy("data/foo", "data/foobar")
    assert storage_dir.cat() == {"data": {"foo": "foo", "foobar": "foo"}}


def test_copy_collection(storage_dir: TmpDir, client: Client):
    """Test simple copy for collection."""
    dir_internals = {"foo": "foo", "sub": {"file": "file"}}
    storage_dir.gen({"data": dir_internals})
    client.copy("data", "data2")

    assert storage_dir.cat() == {"data": dir_internals, "data2": dir_internals}


def test_try_copy_file_when_destination_already_exists(
    storage_dir: TmpDir, client: Client
):
    """Try copying a resource to a destination that's already mapped/exists."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    with pytest.raises(ResourceAlreadyExists) as exc_info:
        client.copy("data/foo", "data/bar")

    assert str(exc_info.value) == "The resource data/bar already exists"
    assert exc_info.value.path == "data/bar"


def test_try_copy_resource_that_does_not_exist(
    storage_dir: TmpDir, client: Client
):
    """Try to copy a resource that does not exist."""
    with pytest.raises(ResourceNotFound) as exc_info:
        client.copy("data", "data2")

    assert storage_dir.cat() == {}
    assert str(exc_info.value) == (
        "The resource data could not be found in the server"
    )
    assert exc_info.value.path == "data"


@pytest.mark.parametrize("from_path", ["data", "data/foo"])
def test_try_copy_to_destination_parent_does_not_exist(
    storage_dir: TmpDir, client: Client, from_path: str
):
    """Try to copy a path to a destination whose parent does not exist yet."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})

    with pytest.raises(ResourceConflict) as exc_info:
        client.copy(from_path, "data3/bar")

    assert str(exc_info.value) == (
        "there was a conflict when trying to copy the resource"
    )
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}


def test_copy_overwrite_file(storage_dir: TmpDir, client: Client):
    """Test copy file with overwrite."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    client.copy("/data/foo", "data/bar/", overwrite=True)
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "foo"}}


def test_copy_collection_overwrite(storage_dir: TmpDir, client: Client):
    """Test copy a collection with overwrite."""
    dir_internals = {"foo": "foo", "sub": {"file": "file"}}
    storage_dir.gen({"data": dir_internals, "data2": {"foo": "foo"}})

    client.copy("/data", "/data2/", overwrite=True)
    assert storage_dir.cat() == {"data": dir_internals, "data2": dir_internals}


def test_move_file(storage_dir: TmpDir, client: Client):
    """Test simple file move."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    client.move("data/foo", "data/foobar")
    assert storage_dir.cat() == {"data": {"bar": "bar", "foobar": "foo"}}


def test_move_collection(storage_dir: TmpDir, client: Client):
    """Test simple collection move."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    client.move("data", "data2")
    assert storage_dir.cat() == {"data2": {"foo": "foo", "bar": "bar"}}


def test_try_move_resource_that_does_not_exist(
    storage_dir: TmpDir, client: Client
):
    """Test trying to move a resource that does not exist at all."""
    with pytest.raises(ResourceNotFound) as exc_info:
        client.move("data", "data2")

    assert storage_dir.cat() == {}
    assert str(exc_info.value) == (
        "The resource data could not be found in the server"
    )

    assert exc_info.value.path == "data"


def test_move_file_dest_exists_already(storage_dir: TmpDir, client: Client):
    """Test moving file to a destination that already exists."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    with pytest.raises(ResourceAlreadyExists) as exc_info:
        client.move("data/foo", "data/bar")
    assert str(exc_info.value) == ("The resource data/bar already exists")
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}


def test_move_collection_dest_exists_already(
    storage_dir: TmpDir, client: Client
):
    """Test moving a collection to a destination that already exists."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}, "data2": {}})
    with pytest.raises(ResourceAlreadyExists) as exc_info:
        client.move("data", "data2")

    assert str(exc_info.value) == ("The resource data2 already exists")
    assert storage_dir.cat() == {
        "data": {"foo": "foo", "bar": "bar"},
        "data2": {},
    }


def test_move_file_with_overwrite(storage_dir: TmpDir, client: Client):
    """Test moving a file to a dest. that already exists and overwrite."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    client.move("data/foo", "data/bar", overwrite=True)
    assert storage_dir.cat() == {"data": {"bar": "foo"}}


def test_move_collection_with_overwrite(storage_dir: TmpDir, client: Client):
    """Test moving a coll. to a dest. that already exists and overwrite."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}, "data2": {}})
    client.move("data", "data2", overwrite=True)
    assert storage_dir.cat() == {"data2": {"foo": "foo", "bar": "bar"}}


@pytest.mark.parametrize("from_path", ["data", "data/foo"])
def test_move_to_a_dest_whose_parent_does_not_exist(
    storage_dir: TmpDir, client: Client, from_path: str
):
    """Test moving a resource to a dest. whose parent don't exists yet."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})

    with pytest.raises(ResourceConflict) as exc_info:
        client.move(from_path, "data3/bar")

    assert str(exc_info.value) == (
        "there was a conflict when trying to move the resource"
    )
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}


@pytest.mark.parametrize(
    "lock_path, move_from",
    [
        # Lock a collection resource and try to move it.
        ("data", "data"),
        # Lock a non-collection resource and try to move it.
        ("data/foo", "data/foo"),
        # Lock a collection resource and try to move it's member resource.
        ("data", "data/foo"),
        # Lock a non-collection resource and try to move its parent collection.
        ("data/foo", "data"),
        # Lock a destination collection and try to move a collection into it.
        ("data2", "data"),
        # Lock a destination collection and try to move a non-collection to it.
        ("data2", "data/foo"),
        # Lock a destination non-collection and try to move a collection to it.
        ("data2/foobar", "data"),
        # Lock a destination collection and try to move a non-collection to it.
        ("data2/foobar", "data/foo"),
    ],
)
def test_try_moving_a_resource_locked(
    storage_dir: TmpDir,
    client: Client,
    move_from: str,
    lock_path: str,
):
    """Test trying to move a resource that's locked.

    (completely or partially, in src or dest)
    """
    storage_dir.gen(
        {"data": {"foo": "foo", "bar": "bar"}, "data2": {"foobar": "foobar"}}
    )
    lock_resource(client, lock_path)

    with pytest.raises(ResourceLocked) as exc_info:
        client.move(move_from, "data2")

    assert (
        str(exc_info.value)
        == "the source or the destination resource is locked"
    )

    # should not have been moved at all
    assert storage_dir.cat() == {
        "data": {"foo": "foo", "bar": "bar"},
        "data2": {"foobar": "foobar"},
    }


@pytest.mark.parametrize("method", ["copy", "move"])
def test_transfer_multistatus_failure(client: Client, method: str):
    """Test multistatus error in move and copy.

    Since it's hard to reproduce them with the current wsgidav, went with the
    xml raw response content to test with.
    """
    from httpx import Request, Response

    url = "http://www.example.com/container/"
    request = Request(HTTPMethod.MOVE, url)
    response = Response(
        status_code=207,
        request=request,
        content="""<?xml version="1.0" encoding="utf-8" ?>
        <d:multistatus xmlns:d='DAV:'>
            <d:response>
                <d:href>/othercontainer/C2/</d:href>
                <d:status>HTTP/1.1 423 Locked</d:status>
                <d:error><d:lock-token-submitted/></d:error>
            </d:response>
        </d:multistatus>""",
    )

    client.http.request = MagicMock(  # type: ignore[assignment]
        return_value=response
    )
    func = getattr(client, method)
    with pytest.raises(MultiStatusError) as exc_info:
        func("/container", "/othercontainer")

    assert str(exc_info.value) == "The resource /othercontainer/C2/ is locked"


@pytest.mark.parametrize("method", ["copy", "move"])
def test_transfer_forbidden_operations(
    client: Client, server_address: URL, method: str
):
    """Test that copy and move handle forbidden operations.

    This might happen due to number of issues, but mainly it's because
    the source and destination resource are the same.
    """
    from httpx import Request, Response

    url = server_address.join("container")
    request = Request(HTTPMethod.MKCOL, url)
    response = Response(status_code=403, request=request)

    client.http.request = MagicMock(  # type: ignore[assignment]
        return_value=response
    )

    func = getattr(client, method)
    with pytest.raises(ForbiddenOperation) as exc_info:
        func("container", "othercontainer")

    assert str(exc_info.value) == (
        "the source and the destination could be the same"
    )


def test_mkdir(storage_dir: TmpDir, client: Client):
    """Test simple mkdir creation."""
    client.mkdir("data")
    assert storage_dir.cat() == {"data": {}}


def test_mkdir_but_parent_collection_not_exist(
    storage_dir: TmpDir, client: Client
):
    """Test creating a collection but parent collection does not exist."""
    with pytest.raises(ResourceConflict) as exc_info:
        client.mkdir("data/sub")

    assert storage_dir.cat() == {}
    assert str(exc_info.value) == "parent of the collection does not exist"


def test_mkdir_collection_already_exists(storage_dir: TmpDir, client: Client):
    """Test trying to create an already-existing collection."""
    storage_dir.gen({"data": {"foo": "foo"}})
    with pytest.raises(ResourceAlreadyExists) as exc_info:
        client.mkdir("data")

    assert storage_dir.cat() == {"data": {"foo": "foo"}}
    assert str(exc_info.value) == "The resource data already exists"
    assert exc_info.value.path == "data"


def test_mkdir_forbidden_operations(client: Client, server_address: URL):
    """Test forbidden operations for mkdir.

    Notes:
        This indicates at least one of two conditions (from rfc4918):
        1) the server does not allow the creation of collections at the
            given location in its URL namespace, or
        2) the parent collection of the Request-URI exists but cannot accept
            members
    """
    from httpx import Request, Response

    url = server_address.join("container")
    request = Request(HTTPMethod.MKCOL, url)
    response = Response(status_code=403, request=request)

    client.http.request = MagicMock(  # type: ignore[assignment]
        return_value=response
    )
    with pytest.raises(ForbiddenOperation) as exc_info:
        client.mkdir("container")

    assert str(exc_info.value) == (
        "the server does not allow creation in the namespace"
        "or cannot accept members"
    )


@pytest.mark.parametrize(
    "path", ["collections", "/collections", "/collections", "/collections/"]
)
def test_mkdir_sends_a_trailing_slash(path: str):
    """Test that mkdir sends a request to the url with a trailing slash.

    See: https://github.com/skshetry/webdav4/issues/55
    """
    from httpx import Request, Response

    response = Response(200, request=Request(Method.MKCOL, "url"))
    client = Client(
        "http://example.org", http_client=mock.MagicMock(return_value=response)
    )
    with mock.patch.object(client.http, "request", return_value=response) as m:
        client.mkdir(path)
        assert m.call_args == mock.call(
            "MKCOL", URL("http://example.org/collections/")
        )


def test_remove_collection(storage_dir: TmpDir, client: Client):
    """Test trying to remove a collection resource."""
    storage_dir.gen({"data": {"foo": "foo"}})
    client.remove("data")
    assert storage_dir.cat() == {}


def test_remove_non_collection(storage_dir: TmpDir, client: Client):
    """Test trying to remove a non-collection resource."""
    storage_dir.gen({"data": {"foo": "foo"}})
    client.remove("data/foo")
    assert storage_dir.cat() == {"data": {}}


def test_remove_not_existing_resource(client: Client):
    """Test trying to remove a resource that does not exist."""
    with pytest.raises(ResourceNotFound) as exc_info:
        client.remove("data")

    assert (
        str(exc_info.value)
        == "The resource data could not be found in the server"
    )
    assert exc_info.value.path == "data"


def test_try_remove_locked_resource_non_coll(
    storage_dir: TmpDir, client: Client
):
    """Test trying to remove a resource that is locked."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    lock_resource(client, "data/foo")

    with pytest.raises(ResourceLocked) as exc_info:
        client.remove("data/foo")

    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}
    assert str(exc_info.value) == "the resource is locked"


def test_try_remove_locked_resource_coll(
    storage_dir: TmpDir, client: Client, server_address: URL
):
    """Test trying to remove a resource that is locked."""
    storage_dir.gen({"data": {"foo": "foo", "bar": "bar"}})
    lock_resource(client, "data")

    with pytest.raises(MultiStatusError) as exc_info:
        client.remove("data")

    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}

    statuses = {
        client.join_url("/data").path + "/": "Locked",
        client.join_url("/data/bar").path: "Locked",
        client.join_url("/data/foo").path: "Locked",
    }
    assert str(exc_info.value) == (
        "multiple errors received: " + str(statuses)
    )


@pytest.mark.parametrize(
    "structure, path",
    [
        ({"data": {"foo": "foo", "bar": "bar"}}, "/data"),
        ({"data": {"foo": "foo", "bar": "bar"}}, "/data/foo"),
    ],
)
def test_client_propfind(
    structure: Dict[str, Any],
    path: str,
    storage_dir: "TmpDir",
    client: Client,
):
    """Test http client's propfind response."""
    storage_dir.gen(structure)
    client.ls(path, detail=False)


def test_open(storage_dir: TmpDir, client: Client):
    """Test opening a remote file from webdav opened in text mode."""
    text1 = "0123456789" * (DEFAULT_BUFFER_SIZE // 10 + 1)
    storage_dir.gen("sample.txt", text1 * 2)

    with client.open("sample.txt", mode="r") as f:
        assert not f.closed
        assert not f.isatty()
        assert f.readable()
        assert not f.writable()
        assert f.seekable()
        assert f.tell() == 0
        assert f.read(len(text1)) == text1
        assert f.tell() == len(text1)
        assert f.seek(10) == 10
        assert f.read(len(text1) - 10) == text1[10:]
        assert f.tell() == len(text1)
        assert f.seek(len(text1) * 2) == len(text1) * 2
        assert f.read() == ""
        assert f.seek(0) == 0
    assert f.closed
    f.close()

    text2 = "0123456789\n"
    storage_dir.gen("sample2.txt", text2 * 10)

    with client.open("sample2.txt", mode="r") as f:
        assert f.readline() == text2
        assert f.readlines() == [text2] * 9
    assert f.closed
    f.close()


def test_open_binary(storage_dir: TmpDir, client: Client):
    """Test file object opened in binary mode with client."""
    text1 = b"0123456789" * (DEFAULT_BUFFER_SIZE // 10 + 1)
    storage_dir.gen("sample.txt", text1 * 2)
    with client.open("sample.txt", mode="rb") as f:
        assert not f.closed
        assert not f.isatty()
        assert f.readable()
        assert not f.writable()
        assert f.seekable()
        assert f.tell() == 0

        assert f.read(len(text1)) == text1
        assert f.tell() == len(text1)

        assert f.seek(10) == 10
        assert f.read(len(text1) - 10) == text1[10:]
        assert f.tell() == len(text1)

        assert f.seek(len(text1), 1) == len(text1) * 2
        assert f.read() == b""
        assert f.seek(-len(text1), 1) == len(text1)

        assert f.readall() == text1  # type: ignore

        assert f.seek(0) == 0
        buff = bytearray(5)
        assert f.readinto(buff) == 5  # type: ignore
        assert bytes(buff) == text1[:5]
        length = f.readinto1(buff)  # type: ignore
        assert bytes(buff) == text1[5 : 5 + length]

        assert f.read1(10)  # type: ignore
        assert f.seek(-10, 2) == f.tell() == len(text1) * 2 - 10
        assert f.read() == text1[-10:]
    assert f.closed
    f.close()

    with client.open("sample.txt", mode="rb") as f:
        with pytest.raises(ValueError):
            f.seek(-10)
        with pytest.raises(ValueError):
            f.seek(10, 3)
        with mock.patch.object(
            type(f), "size", new_callable=mock.PropertyMock
        ) as m:
            m.return_value = None
            with pytest.raises(ValueError):
                f.seek(-10, 2)

    assert f.closed

    text2 = b"0123456789\n"
    storage_dir.gen("sample2.txt", text2 * 10)

    with client.open("sample2.txt", mode="rb") as f:
        assert f.readline() == text2
        assert f.readlines() == [text2] * 9
    assert f.closed
    f.close()


def test_feature_detection_util():
    """Test parsing of response in FeatureDetection util."""
    from httpx import Response

    fd = FeatureDetection()
    assert fd.supports_ranges is False
    assert not fd.dav_compliances

    response = Response(
        HTTPStatus.OK, headers={"DAV": "1,2,3", "Accept-Range": "bytes"}
    )
    fd = FeatureDetection(response)
    assert fd.supports_ranges is False
    assert fd.dav_compliances


def test_download_fobj(storage_dir: TmpDir, client: Client):
    """Test downloading a resource to a file object."""
    storage_dir.gen({"data": {"foo": "foo"}})
    buff = BytesIO()
    client.download_fileobj("/data/foo", buff)
    assert buff.getvalue() == b"foo"


def test_download_file(tmp_path: Path, storage_dir: TmpDir, client: Client):
    """Test downloading a remote resource to a local file."""
    storage_dir.gen({"data": {"foo": "foo"}})
    file_path = tmp_path / "foo.txt"
    client.download_file("/data/foo", file_path)
    assert file_path.read_text()


def test_try_download_directory(
    tmp_path: Path, storage_dir: TmpDir, client: Client
):
    """Test downloading a remote resource to a local file."""
    storage_dir.gen({"data": {"foo": "foo"}})

    with pytest.raises(IsACollectionError) as exc_info:
        client.download_file("/data", tmp_path / "foo")

    assert (
        str(exc_info.value)
        == "/data is a collection. Cannot open a collection"
    )


def test_try_downloading_not_existing_resource(tmp_path: Path, client: Client):
    """Test trying to download a resource that does not exist at all."""
    file_path = tmp_path / "foo"
    with pytest.raises(ResourceNotFound) as exc_info:
        client.download_file("foo", file_path)

    assert (
        str(exc_info.value)
        == "The resource foo could not be found in the server"
    )


def test_open_file(storage_dir: TmpDir, client: Client):
    """Testing opening a remote url to a file-like object."""
    storage_dir.gen("foo", "foo")
    with client.open("foo") as f:
        assert f.read() == "foo"
        assert f.read() == ""
        assert f.read() == ""
        assert f.read(0) == ""

    with client.open("foo", mode="rb") as f:
        assert f.read() == b"foo"
        assert f.read() == b""
        assert f.read() == b""
        assert f.read(0) == b""

    with client.open("foo", mode="rb") as f:
        assert f.readall() == b"foo"  # type: ignore

    with client.open("foo", mode="rb") as f:
        b = bytearray(range(10))
        assert f.readinto(b) == 3  # type: ignore
        assert b[:3] == bytearray(b"foo")


def test_raising_insufficient_storage():
    """Test that insufficient storage is raised."""
    from httpx import Request, Response

    client = Client("https://example.org")
    url = client.join_url("test")
    req = Request(HTTPMethod.COPY, url)
    resp = Response(status_code=507, request=req)

    client.http.request = MagicMock(  # type: ignore[assignment]
        return_value=resp
    )
    with pytest.raises(InsufficientStorage) as exc_info:
        client.request("copy", "test")

    assert str(exc_info.value) == "Insufficient Storage on the server"


def test_upload_fobj(storage_dir: TmpDir, client: Client):
    """Test uploading a resource from a file object."""
    buff = BytesIO()
    buff.write(b"foo")

    # to decide size of file in bytesio, we need to seek to the start.
    # otherwise, it'll fail to determine size of file object
    # which will then try using chunked upload which is not supported by
    # cheroot server which we use for testing.
    buff.seek(0)

    client.upload_fileobj(buff, "foo")

    assert client.exists("/foo")
    assert client.isfile("/foo")
    assert not client.isdir("/foo")
    assert storage_dir.cat() == {"foo": "foo"}


def test_upload_fobj_size_hints(client: Client):
    """Test uploading file object with provided size as hints.

    This is used to choose between chunked and non-chunked transfers.
    """
    from .test_callback import ReadWrapper

    with patch.object(client, "request") as m:
        client.upload_fileobj(BytesIO(b"foobar"), "foobar", size=3)
        assert m.call_args[1]["headers"] == {"Content-Length": "3"}

        client.upload_fileobj(BytesIO(b"foobar"), "foobar")
        assert m.call_args[1]["headers"] == {"Content-Length": "6"}

        wrapped = ReadWrapper(BytesIO(b"foobar"))  # type: ignore
        client.upload_fileobj(wrapped, "foobar")  # type: ignore
        assert m.call_args[1]["headers"] is None

        wrapped = ReadWrapper(BytesIO(b"foobar"))  # type: ignore
        client.upload_fileobj(wrapped, "foobar", size=6)  # type: ignore
        assert m.call_args[1]["headers"] == {"Content-Length": "6"}


def test_upload_file(tmp_path: Path, storage_dir: TmpDir, client: Client):
    """Test uploading a local file to a remote."""
    file_path = tmp_path / "foo.txt"
    file_path.write_text("foo")
    client.upload_file(file_path, "foo")

    assert client.exists("/foo")
    assert client.isfile("/foo")
    assert not client.isdir("/foo")
    assert storage_dir.cat() == {"foo": "foo"}


def test_try_upload_file_that_already_exists(
    tmp_path: Path, storage_dir: TmpDir, client: Client
):
    """Test uploading a local file to a remote path that is already mapped.."""
    storage_dir.gen({"foo": "foo"})

    file_path = tmp_path / "foobar"
    file_path.write_text("foobar")

    with pytest.raises(ResourceAlreadyExists) as exc_info:
        client.upload_file(file_path, "foo")

    assert str(exc_info.value) == "The resource foo already exists"


def test_upload_file_with_overwrite(
    tmp_path: Path, storage_dir: TmpDir, client: Client
):
    """Test overwriting an already existing file when uploading."""
    storage_dir.gen({"foo": "foo"})

    file_path = tmp_path / "foobar"
    file_path.write_text("foobar")

    client.upload_file(file_path, "foo", overwrite=True)
    assert storage_dir.cat() == {"foo": "foobar"}


def test_isdir_isfile(storage_dir: TmpDir, client: Client):
    """Test isdir and isfile checks."""
    storage_dir.gen({"data": {"foo": "foo"}})

    assert client.isdir("/data/")
    assert not client.isfile("/data/")

    assert client.isfile("/data/foo")
    assert not client.isdir("/data/foo")

    with pytest.raises(ResourceNotFound) as exc_info:
        client.isdir("/data/file")

    assert exc_info.value.path == "/data/file"
    assert (
        str(exc_info.value)
        == "The resource /data/file could not be found in the server"
    )

    with pytest.raises(ResourceNotFound) as exc_info:
        client.isdir("/data/file")

    assert exc_info.value.path == "/data/file"
    assert (
        str(exc_info.value)
        == "The resource /data/file could not be found in the server"
    )


def test_exists(storage_dir: TmpDir, client: Client):
    """Test exists."""
    storage_dir.gen({"data": {"foo": "foo"}})

    assert client.exists("/data/")
    assert client.exists("/data/foo")
    assert not client.exists("/data/bar")


def test_attributes(storage_dir: TmpDir, client: Client):
    """Test APIs that provides attributes of the resource."""
    storage_dir.gen({"data": {"foo": "foo"}})

    stat = (storage_dir / "data" / "foo").stat()
    assert client.content_length("/data/foo") == 3
    assert client.content_type("/data/foo") == "application/octet-stream"
    assert client.content_language("/data/foo") == ""
    etag = client.etag("/data/foo")
    assert etag and isinstance(etag, str)
    assert client.modified("/data/foo") == approx_datetime(
        datetime.fromtimestamp(int(stat.st_mtime), tz=timezone.utc)
    )
    assert client.created("/data/foo") == approx_datetime(
        datetime.fromtimestamp(int(stat.st_ctime), tz=timezone.utc)
    )


def test_attributes_dir(storage_dir: TmpDir, client: Client):
    """Test attributes response on a directory."""
    storage_dir.gen({"data": {"foo": "foo"}})

    stat = (storage_dir / "data").stat()
    assert client.content_length("/data/") is None
    assert client.content_type("/data/") == ""
    assert client.content_language("/data/") == ""
    assert client.etag("/data/") is None
    assert client.modified("/data/") == approx_datetime(
        datetime.fromtimestamp(int(stat.st_mtime), tz=timezone.utc)
    )
    assert client.created("/data/") == approx_datetime(
        datetime.fromtimestamp(int(stat.st_ctime), tz=timezone.utc)
    )


def test_attributes_file_not_exist(client: Client):
    """Test attributes' API throws exception if the file does not exist."""

    def assert_raises(func: Callable[[str], Any]) -> None:
        with pytest.raises(ResourceNotFound):
            func("data")

    assert_raises(client.content_length)
    assert_raises(client.content_type)
    assert_raises(client.content_language)
    assert_raises(client.etag)
    assert_raises(client.modified)
    assert_raises(client.created)


def test_client_bad_gateway_error(client: Client, server_address: URL):
    """Test bad gateway errors.

    This are raised mostly on proxy errors. But on webdav, it might
    also happen if the destination for the `COPY` or `MOVE` operations
    are in remote server and the remote server refused the request or
    is unavailable to address the request.
    """
    from httpx import Request, Response

    url = server_address.join("container")
    request = Request(HTTPMethod.PROPFIND, url)
    response = Response(status_code=502, request=request)

    client.http.request = MagicMock(  # type: ignore[assignment]
        return_value=response
    )
    with pytest.raises(BadGatewayError) as exc_info:
        client.request(HTTPMethod.PROPFIND, "/container")

    assert (
        str(exc_info.value) == "The destination server may have refused"
        " to accept the resource"
    )


def test_options(client: Client):
    """Test features of the webdav server.

    Note that, right now, we are only testing with `WsgiDav` server.
    If this changes, we will need to figure out how to change this
    expectation here.
    """
    assert client.options() == {"1", "2"}


def test_feature_detection(client: Client):
    """Test feature detection in the Client."""
    assert client.detected_features
    assert client._detected_features  # pylint: disable=protected-access
    assert client.detected_features.dav_compliances

    client._detected_features = None  # pylint: disable=protected-access
    # should not fail at all even if auth is invalid
    client.http.auth = ("invalid", "invalid")  # type: ignore
    assert client.detected_features.dav_compliances
    assert client.detected_features


def test_auth_errors(server_address: URL):
    """Test that auth errors are raised as http error.

    This test serves another purpose as well: it is testing that
    unhandled exceptions are re-raising `HTTPError`.
    """
    client = Client(server_address, ("wrong", "password"))
    expected = "received 401 (Not Authorized)"
    path = "resource-that-does-not-exist"
    with pytest.raises(HTTPError) as exc_info:
        client.get_props(path)
    assert str(exc_info.value) == expected

    with pytest.raises(HTTPError) as exc_info:
        client.propfind(path)
    assert str(exc_info.value) == expected

    with pytest.raises(HTTPError) as exc_info:
        client.mkdir(path)
    assert str(exc_info.value) == expected

    with pytest.raises(HTTPError) as exc_info:
        client.remove(path)
    assert str(exc_info.value) == expected

    with pytest.raises(HTTPError) as exc_info:
        client.move(path, "resource2")
    assert str(exc_info.value) == expected

    with pytest.raises(HTTPError) as exc_info:
        client.copy(path, "resource2")
    assert str(exc_info.value) == expected


def test_client_retries(client: Client, server_address: URL):
    """Test that the client retries."""
    from httpx import Request, Response

    url = server_address.join("container1")
    request = Request(HTTPMethod.PROPFIND, url)
    failed_response = Response(status_code=502, request=request)
    success_response = Response(
        status_code=HTTPStatus.OK.value, request=request
    )

    client.http.request = func = MagicMock(  # type: ignore[assignment]
        side_effect=[failed_response, failed_response, success_response]
    )
    client.copy("/container1", "/container2")
    assert func.call_count == 3
